from unittest import TestCase

from musicscore.metronome import Metronome
from musicscore.quarterduration import QuarterDuration
from musurgia.musurgia_exceptions import WrongTreeValueError
from musurgia.tests.utils_for_tests import create_test_timeline_tree
from musurgia.trees.timelinetree import TimelineDuration, TimelineTree


class TimelineDurationTestCase(TestCase):
    def setUp(self):
        self.tld = TimelineDuration(4)
        return super().setUp()

    def test_set_quarter_duration_error(self):
        with self.assertRaises(AttributeError):
            self.tld.quarter_duration = QuarterDuration(4)

    def test_get_quarter_duration(self):
        self.assertEqual(self.tld.metronome.per_minute, 60)
        self.assertEqual(self.tld.get_quarter_duration(), QuarterDuration(4))

    def test_update_quarter_duration(self):
        self.assertEqual(self.tld.get_quarter_duration(), 4)
        self.tld.metronome = Metronome(120)
        self.assertEqual(self.tld.get_quarter_duration(), 8)
        self.tld._set_seconds(5)
        self.assertEqual(self.tld.get_quarter_duration(), 10)
        self.tld.metronome = Metronome(120, 2)
        self.assertEqual(self.tld.get_quarter_duration(), 20)


class TimeLineTreeTestCase(TestCase):
    def test_create_timeline_tree_root(self):
        tlt = TimelineTree(TimelineDuration(2))
        self.assertEqual(tlt.get_duration().calculate_in_seconds(), 2)

    def test_add_child_to_timeline(self):
        root_duration = TimelineDuration(2)
        child_durations = [TimelineDuration(1.5), TimelineDuration(0.5)]
        tlt = TimelineTree(root_duration)
        [tlt.add_child(TimelineTree(d)) for d in child_durations]
        self.assertListEqual(
            [ch.get_duration() for ch in tlt.get_children()], child_durations
        )

    def test_check_timeline_durations(self):
        tlt = TimelineTree(TimelineDuration(2))
        tlt.add_child(TimelineTree(TimelineDuration(1.5)))
        tlt.add_child(TimelineTree(TimelineDuration(0.5)))
        self.assertTrue(tlt.check_tree_values())
        tlt.add_child(TimelineTree(TimelineDuration(1)))
        with self.assertRaises(WrongTreeValueError) as err:
            tlt.check_tree_values()
        expected = "Children of ValuedTree node of position 0 with value 2 have wrong values [Fraction(3, 2), Fraction(1, 2), Fraction(1, 1)] (sum=3)"
        self.assertEqual(str(err.exception), expected)

    def test_timeline_get_value(self):
        tft = create_test_timeline_tree()
        self.assertEqual(tft.get_value(), 60)
        expected = """└── 60.0
    ├── 20.0
    │   ├── 10.0
    │   ├── 2.0
    │   ├── 3.0
    │   └── 5.0
    ├── 10.0
    └── 30.0
        ├── 5.0
        ├── 20.0
        ├── 3.0
        └── 2.0
"""
        self.assertEqual(
            tft.get_tree_representation(key=lambda node: float(node.get_value())),
            expected,
        )

    def test_timeline_update_duration(self):
        tft = create_test_timeline_tree()
        one_child = tft.get_children()[0].get_children()[0]
        self.assertEqual(one_child.get_value(), 10)
        self.assertEqual(one_child.get_duration(), 10)

        with self.assertRaises(AttributeError):
            one_child.value = 15

        with self.assertRaises(AttributeError):
            one_child.duration.value = 15

        with self.assertRaises(AttributeError):
            one_child.duration.seconds = 15

        with self.assertRaises(AttributeError):
            one_child.duration = TimelineDuration(15)

        one_child.update_duration(15)
        self.assertEqual(one_child.get_value(), 15)
        self.assertEqual(one_child.get_duration(), 15)

        expected = """└── 65.0
    ├── 25.0
    │   ├── 15.0
    │   ├── 2.0
    │   ├── 3.0
    │   └── 5.0
    ├── 10.0
    └── 30.0
        ├── 5.0
        ├── 20.0
        ├── 3.0
        └── 2.0
"""
        self.assertEqual(
            tft.get_tree_representation(key=lambda node: float(node.get_value())),
            expected,
        )
        tft.check_tree_values()

        one_child.update_duration(TimelineDuration(5))
        self.assertEqual(one_child.get_value(), 5)
        self.assertEqual(one_child.get_duration(), 5)
        expected = """└── 55.0
    ├── 15.0
    │   ├── 5.0
    │   ├── 2.0
    │   ├── 3.0
    │   └── 5.0
    ├── 10.0
    └── 30.0
        ├── 5.0
        ├── 20.0
        ├── 3.0
        └── 2.0
"""
        self.assertEqual(
            tft.get_tree_representation(key=lambda node: float(node.get_value())),
            expected,
        )

        tft.check_tree_values()



from unittest import TestCase

from musurgia.musurgia_exceptions import (
    ClockWrongSecondsValueError,
    ClockWrongMinutesValueError,
    ClockWrongSecondsTypeError,
    ClockWrongMinutesTypeError,
    ClockWrongHoursTypeError,
)
from musurgia.timing.clock import Clock


class TestClock(TestCase):
    def test_clock_errors(self):
        with self.assertRaises(ClockWrongSecondsValueError):
            Clock(0, 0, 70.0)
        with self.assertRaises(ClockWrongMinutesValueError):
            Clock(0, 70, 0.0)
        with self.assertRaises(ClockWrongSecondsTypeError):
            Clock(0, 0, "30")
        with self.assertRaises(ClockWrongMinutesTypeError):
            Clock(0, 30.0, 0.0)
        with self.assertRaises(ClockWrongHoursTypeError):
            Clock(1.0, 0, 0.0)

    def test_clock_mode(self):
        c = Clock(4, 2, 0.5)
        assert c.get_as_string(mode="hms") == "4:02:00.5"
        assert c.get_as_string(mode="ms") == "02:00.5"
        assert c.get_as_string(mode="msreduced") == "2:0.5"
        assert c.get_as_string(mode="msreduced") == "2:0.5"

        c.set_values(0, 0, 1.5)
        assert c.get_as_string(mode="msreduced") == "1.5"

        c.set_values(4, 2, 1)
        assert c.get_as_string(mode="hms") == "4:02:01.0"
        assert c.get_as_string(mode="ms") == "02:01.0"
        assert c.get_as_string(mode="msreduced") == "2:1.0"
        assert c.get_as_string(mode="msreduced") == "2:1.0"

    def test_get_as_string_round(self):
        c = Clock(4, 2, 1.2568)
        assert c.get_as_string(mode="hms", round_=2) == "4:02:01.26"

    def test_get_and_set_values(self):
        c = Clock(4, 2, 0.5)
        assert c.get_values() == (4, 2, 0.5)
        c.set_values(3, 1, 10.5)
        assert c.get_as_string() == "3:01:10.5"
        c.minutes = 5
        assert c.get_as_string() == "3:05:10.5"

    def test_add_clocks(self):
        c1 = Clock(1, 2, 0.5)
        c2 = Clock(2, 59, 59.5)
        c = c1 + c2
        assert c.get_values() == (4, 2, 0.0)

    def test_subtract_clocks(self):
        c1 = Clock(4, 2, 0.0)
        c2 = Clock(2, 59, 59.5)
        c = c1 - c2
        assert c.get_values() == (1, 2, 0.5)



from musurgia.pdf.line import VerticalLineSegment
from musurgia.pdf.pdf import Pdf
from musurgia.tests.utils_for_tests import PdfTestCase


class TestAbsoluteY(PdfTestCase):
    def setUp(self) -> None:
        self.pdf = Pdf()

    def test_translate(self):
        self.pdf.translate(10, 30)
        self.pdf.translate(20, 10)
        actual = self.pdf.absolute_y
        expected = 40
        self.assertEqual(expected, actual)

    def test_draw_objects(self):
        vls = VerticalLineSegment(length=20)
        vls.draw(self.pdf)
        actual = self.pdf.absolute_y
        expected = 0
        self.assertEqual(expected, actual)

    def test_draw_objects_and_translate(self):
        vls = VerticalLineSegment(length=20)
        self.pdf.translate(10, 30)
        vls.draw(self.pdf)
        self.pdf.translate(20, 10)
        actual = self.pdf.absolute_y
        expected = 40
        self.assertEqual(expected, actual)



from unittest import TestCase

from musurgia.pdf.line import SlaveStraightLine
from musurgia.tests.utils_for_tests import DummyMaster


class TestStraightLine(TestCase):
    def test_is_horizontal_or_vertical(self):
        line = SlaveStraightLine(
            master=DummyMaster(), value="something", mode="h", length=2
        )
        assert line.is_horizontal
        assert not line.is_vertical
        line.mode = "horizontal"
        assert line.is_horizontal
        assert not line.is_vertical
        line.mode = "vertical"
        assert line.is_vertical
        assert not line.is_horizontal
        line.mode = "v"
        assert line.is_vertical
        assert not line.is_horizontal

    def test_get_opposite_mode(self):
        assert SlaveStraightLine.get_opposite_mode("v") == "h"
        assert SlaveStraightLine.get_opposite_mode("vertical") == "horizontal"
        assert SlaveStraightLine.get_opposite_mode("h") == "v"
        assert SlaveStraightLine.get_opposite_mode("horizontal") == "vertical"



from abc import ABC, abstractmethod
from fractions import Fraction
from typing import Any, Iterator, Literal, Union

from musicscore.midi import Midi
from musurgia.magicrandom import MagicRandom
from musurgia.utils import RelativeValueGenerator


class MidiGenerator(ABC):
    def __init__(self, *args: Any, **kwargs: Any):
        super().__init__(*args, **kwargs)

    @abstractmethod
    def __next__(self) -> Midi:
        pass


class OneMidiGenerator(MidiGenerator):
    def __init__(self, midi_value: Union[float, int], *args: Any, **kwargs: Any):
        super().__init__(*args, **kwargs)
        self._midi_value = midi_value

    def __next__(self) -> Midi:
        return Midi(self._midi_value)


class RandomMidiGenerator(MagicRandom, MidiGenerator):
    def __next__(self) -> Midi:
        return Midi(float(super().__next__()))


MicrotoneValueTypes = Literal[2, 4, 8]


class RelativeMidiGenerator(RelativeValueGenerator, MidiGenerator):
    def __init__(self, microtone: MicrotoneValueTypes = 2, *args: Any, **kwargs: Any):
        super().__init__(*args, **kwargs)
        self._microtone: MicrotoneValueTypes
        self.microtone = microtone

    @property
    def microtone(self) -> MicrotoneValueTypes:
        return self._microtone

    @microtone.setter
    def microtone(self, value: MicrotoneValueTypes) -> None:
        if value not in [2, 4, 8]:
            raise ValueError("RelativeMidiGenerator.microtone can only be 2, 4 or 8.")
        self._microtone = value
        self.value_grid = Fraction(1, (value // 2))

    def _midi_generator(self) -> Iterator[Midi]:
        for val in self.get_values():
            yield Midi(float(val))

    def __iter__(self) -> Iterator[Midi]:
        self._generator = self._midi_generator()
        return self

    def __next__(self) -> Midi:
        return next(self._generator)



import re
from pathlib import Path
from types import TracebackType
from typing import Protocol, Any, TYPE_CHECKING, Optional, Union, cast

from fpdf import FPDF

from musurgia.musurgia_types import (
    ConvertibleToFloat,
    check_type,
    PageOrientation,
    PageFormat,
)
from musurgia.pdf.pdfunit import PdfUnit

if TYPE_CHECKING:
    from musurgia.pdf.drawobject import DrawObject  # pragma: no cover

__all__ = ["Pdf"]


def sprintf(fmt: str, *args: Any) -> str:
    return fmt % args


class SavedState:
    def __init__(self, pdf: "Pdf") -> None:
        self.pdf = pdf

    def __enter__(self) -> None:
        self.pdf._push_state()

    def __exit__(
        self,
        exc_type: Optional[type[BaseException]],
        exc_val: Optional[BaseException],
        exc_tb: Optional[TracebackType],
    ) -> None:
        self.pdf._pop_state()


class PdfDrawObjectTranslations:
    def __init__(
        self,
        pdf: "Pdf",
        draw_object: "DrawObject",
        translate_positions: bool = True,
        translate_margins: bool = True,
    ) -> None:
        self.pdf = pdf
        self.draw_object = draw_object
        self.translate_positions = translate_positions
        self.translate_margins = translate_margins

    def __enter__(self) -> None:
        self.pdf._push_state()
        if self.translate_positions:
            self.pdf.translate(self.draw_object.relative_x, self.draw_object.relative_y)
        if self.translate_margins:
            self.pdf.translate(
                self.draw_object.left_margin, self.draw_object.top_margin
            )

    def __exit__(
        self,
        exc_type: Optional[type[BaseException]],
        exc_val: Optional[BaseException],
        exc_tb: Optional[TracebackType],
    ) -> None:
        self.pdf._pop_state()


class HasOutProtocol(Protocol):
    def _out(self, s: str) -> None: ...


class Pdf(FPDF, HasOutProtocol):
    """
    Child of fpdf.FPDF class
    """

    def __init__(
        self,
        r_margin: ConvertibleToFloat = 10,
        t_margin: ConvertibleToFloat = 10,
        l_margin: ConvertibleToFloat = 10,
        b_margin: ConvertibleToFloat = 10,
        *args: Any,
        **kwargs: Any,
    ):
        super().__init__(*args, **kwargs)
        self.r_margin = float(r_margin)
        self.t_margin = float(t_margin)
        self.l_margin = float(l_margin)
        self.b_margin = float(b_margin)
        self._absolute_positions: dict[int, list[float]] = {}
        self.add_page()

        self._state: list[str] = []

        self.set_font("Helvetica", "", 10)

    # private

    def _pop_state(self) -> None:
        # https://opensource.adobe.com/dc-acrobat-sdk-docs/pdfstandards/pdfreference1.7old.pdf
        self._state.pop()
        self._out(sprintf("Q\n"))

    def _push_state(self) -> None:
        # https://opensource.adobe.com/dc-acrobat-sdk-docs/pdfstandards/pdfreference1.7old.pdf
        self._state.append("push")
        self._out(sprintf("q\n"))

    # public properties
    @property
    def absolute_positions(self) -> dict[int, list[float]]:
        return self._absolute_positions

    @property
    def absolute_x(self) -> float:
        return self._absolute_positions[self.page][0]

    @property
    def absolute_y(self) -> float:
        return self._absolute_positions[self.page][1]

    @property
    def k(self) -> float:
        return PdfUnit.get_k()

    @k.setter
    def k(self, val: float) -> None:
        pass

    def add_page(
        self,
        orientation: PageOrientation = "",
        format: Union[PageFormat, tuple[float, float]] = "",
        same: bool = False,
        duration: float = 0.0,
        transition: Optional[Any] = None,
        label_style: Optional[Any] = None,
        label_prefix: Optional[str] = None,
        label_start: Optional[int] = None,
    ) -> None:
        check_type(
            orientation,
            "PageOrientation",
            class_name="PageOrientation",
            method_name="add_page",
            argument_name="orientation",
        )
        if not isinstance(format, tuple):
            check_type(
                format,
                "PageFormat",
                class_name="PageOrientation",
                method_name="add_page",
                argument_name="format",
            )
        check_type(
            same,
            bool,
            class_name="PageOrientation",
            method_name="add_page",
            argument_name="same",
        )
        check_type(
            duration,
            float,
            class_name="PageOrientation",
            method_name="add_page",
            argument_name="duration",
        )

        if orientation == "":
            orientation = cast(PageOrientation, self.cur_orientation)
        super().add_page(
            orientation=orientation,
            format=format,
            same=same,
            duration=duration,
            transition=transition,
            label_style=label_style,
            label_prefix=label_prefix,
            label_start=label_start,
        )
        self._absolute_positions[self.page] = [0, 0.0]

    def clip_rect(
        self,
        x: ConvertibleToFloat,
        y: ConvertibleToFloat,
        w: ConvertibleToFloat,
        h: ConvertibleToFloat,
    ) -> None:
        # https://opensource.adobe.com/dc-acrobat-sdk-docs/pdfstandards/pdfreference1.7old.pdf
        check_type(
            x,
            "ConvertibleToFloat",
            class_name=self.__class__.__name__,
            method_name="clip_rect",
            argument_name="x",
        )
        check_type(
            y,
            "ConvertibleToFloat",
            class_name=self.__class__.__name__,
            method_name="clip_rect",
            argument_name="y",
        )
        check_type(
            w,
            "ConvertibleToFloat",
            class_name=self.__class__.__name__,
            method_name="clip_rect",
            argument_name="w",
        )
        check_type(
            h,
            "ConvertibleToFloat",
            class_name=self.__class__.__name__,
            method_name="clip_rect",
            argument_name="h",
        )
        x = float(x)
        y = float(y)
        w = float(w)
        h = float(h)
        self._out(
            sprintf(
                "%.2f %.2f %.2f %.2f re W n",
                x * self.k,
                (self.h - y) * self.k,
                w * self.k,
                -h * self.k,
            )
        )

    def pdf_draw_object_translate(
        self, draw_object: "DrawObject", **kwargs: Any
    ) -> PdfDrawObjectTranslations:
        return PdfDrawObjectTranslations(self, draw_object=draw_object, **kwargs)

    def reset_font(self) -> None:
        # https://opensource.adobe.com/dc-acrobat-sdk-docs/pdfstandards/pdfreference1.7old.pdf
        self._out(
            sprintf(
                "BT /F%d %.2f Tf ET",
                self.current_font.i,  # type: ignore
                self.font_size_pt,
            )
        )

    def reset_position(self) -> None:
        self.translate(-self.absolute_x, -self.absolute_y)

    def saved_state(self) -> SavedState:
        return SavedState(self)

    def translate(self, dx: ConvertibleToFloat, dy: ConvertibleToFloat) -> None:
        check_type(
            dx,
            "ConvertibleToFloat",
            class_name=self.__class__.__name__,
            method_name="translate",
            argument_name="dx",
        )
        check_type(
            dy,
            "ConvertibleToFloat",
            class_name=self.__class__.__name__,
            method_name="translate",
            argument_name="dy",
        )
        dx, dy = float(dx), float(dy)
        # https://opensource.adobe.com/dc-acrobat-sdk-docs/pdfstandards/pdfreference1.7old.pdf
        if not self._state:
            self._absolute_positions[self.page][0] += dx
            self._absolute_positions[self.page][1] += dy
        dx, dy = dx * self.k, dy * self.k
        self._out(sprintf("1.0 0.0 0.0 1.0 %.2F %.2F cm", dx, -dy))

    def translate_page_margins(self) -> None:
        self.translate(self.l_margin, self.t_margin)

    def write_to_path(self, path: Path) -> None:
        buffer = self.output()
        new_buffer = re.sub(
            rb"CreationDate \(D:[0-9]{14}Z[0-9]{2}\'[0-9]{2}\'\)",
            b"CreationDate (none)",
            bytes(buffer),
        )
        new_buffer = bytearray(
            re.sub(rb"ID \[<([0-9A-F]{32})><\1>]", b"ID [<1><2>]", new_buffer)
        )
        path.write_bytes(new_buffer)



from typing import Any, Optional


# arithmetic progression
class ArithmeticProgressionException(Exception):
    pass


class DAndSError(ArithmeticProgressionException):
    def __init__(self, *args: Any):
        msg = "you cannot set both d an s!"
        super().__init__(msg, *args)


# matrix


class MatrixIndexException(Exception):
    pass


class MatrixIndexControllerException(Exception):
    pass


class MatrixIndexOutOfRangeError(MatrixIndexException, ValueError):
    pass


class MatrixIndexEndOfRowError(MatrixIndexException, StopIteration):
    pass


class MatrixIndexEndOfMatrixError(MatrixIndexEndOfRowError, StopIteration):
    pass


class MatrixIndexControllerReadingDirectionError(MatrixIndexControllerException):
    def __init__(self, *args: Any):
        super().__init__(
            "MatrixIndexController.get_next_in_row() works only for reading_direction horizontal.",
            *args,
        )


class MatrixException(Exception):
    pass


class MatrixIsEmptyError(MatrixException):
    def __init__(self) -> None:
        msg = "Matrix is empty!"
        super().__init__(msg)


class SquareMatrixException(MatrixException):
    pass


class SquareMatrixDataError(SquareMatrixException):
    pass


# permutation order matrix


class PermutationOrderMatrixException(MatrixException):
    pass


class PermutationOrderMatrixDataError(PermutationOrderMatrixException):
    pass


# permutation order


class PermutationOrderException(Exception):
    pass


class PermutationOrderError(PermutationOrderException):
    pass


class PermutationOrderTypeError(PermutationOrderError, TypeError):
    pass


class PermutationOrderValueError(PermutationOrderError, ValueError):
    pass


# permutation index


class PermutationIndexCalculatorException(Exception):
    pass


class PermutationIndexCalculaterNoParentIndexError(
    PermutationIndexCalculatorException, ValueError
):
    pass


# fractal timline tree


class FractalTimelineTreeException(Exception):
    pass


class FractalTimelineTreePermutationOrderError(FractalTimelineTreeException):
    pass


class FractalTimelineTreePermutationIndexError(
    FractalTimelineTreeException, ValueError
):
    pass


class FractalTimelineTreeSetMainPermutationOrderFirstError(
    FractalTimelineTreeException, ValueError
):
    msg = "set root's main_permutation_order first"

    def __init__(self, msg: Optional[str] = None):
        if msg is None:
            msg = self.msg
        super().__init__(msg)


class FractalTimelineTreeMergeWrongValuesError(
    FractalTimelineTreeException, ValueError
):
    pass


class FractalTimelineTreeHasNoChildrenError(FractalTimelineTreeException):
    pass


class FractalTimelineTreeHasChildrenError(FractalTimelineTreeException):
    pass


class FractalTimelineTreeNoneRootCannotSetMainPermutationOrderError(
    FractalTimelineTreeException
):
    pass


# pdf


class PdfException(Exception):
    pass


class PdfAttributeError(PdfException, AttributeError):
    pass


# pdf draw object margins


class MarginedObjectException(Exception):
    pass


class MarginNotSettableError(MarginedObjectException):
    pass


# pdf draw object positions


class PositionedObjectException(AttributeError):
    pass


class RelativePositionNotSettableError(PositionedObjectException):
    pass


class RelativeXNotSettableError(RelativePositionNotSettableError):
    pass


class RelativeYNotSettableError(RelativePositionNotSettableError):
    pass


class DrawObjectInContainerHasNegativePositionError(PositionedObjectException):
    pass


# pdf segmented line
class SegmentedLineException(Exception):
    pass


class SegmentedLineSegmentHasMarginsError(SegmentedLineException, AttributeError):
    pass


class SegmentedLineLengthsCannotBeSetError(SegmentedLineException):
    pass


# pdf ruler


class RulerException(SegmentedLineException):
    pass


class RulerCannotSetLengthsError(RulerException, AttributeError):
    pass


class RulerLengthNotPositiveError(RulerException, AttributeError):
    pass


# time ruler


class TimeRulerException(RulerException):
    pass


class TimeRulerCannotSetLength(TimeRulerException, AttributeError):
    pass


# clock
class ClockException(Exception):
    pass


class ClockWrongSecondsValueError(ClockException, ValueError):
    pass


class ClockWrongMinutesValueError(ClockException, ValueError):
    pass


class ClockWrongSecondsTypeError(ClockException, TypeError):
    pass


class ClockWrongMinutesTypeError(ClockException, TypeError):
    pass


class ClockWrongHoursTypeError(ClockException, TypeError):
    pass


# valuedtree


class ValuedTreeException(Exception):
    pass


class WrongTreeValueError(ValuedTreeException, ValueError):
    pass


class WrongTreeValueWarning(Warning):
    pass



from unittest import TestCase

from musurgia.trees.fractaltimelinetree import FractalTimelineTree
from musurgia.matrix.matrix import PermutationOrderMatrix
from musurgia.musurgia_exceptions import (
    FractalTimelineTreeSetMainPermutationOrderFirstError,
)
from musurgia.trees.timelinetree import TimelineDuration


class TestFractalTreeInit(TestCase):
    def test_init(self):
        with self.assertRaises(TypeError):
            FractalTimelineTree()
        with self.assertRaises(TypeError):
            FractalTimelineTree(duration=TimelineDuration(10))
        with self.assertRaises(TypeError):
            FractalTimelineTree(proportions=(1, 2, 3))
        ft = FractalTimelineTree(duration=TimelineDuration(10), proportions=(1, 2, 3))
        with self.assertRaises(FractalTimelineTreeSetMainPermutationOrderFirstError):
            ft.get_permutation_order_matrix()

    def test_init_creates_matrix(self):
        ft = FractalTimelineTree(
            duration=TimelineDuration(10),
            proportions=(1, 2, 3),
            main_permutation_order=(3, 1, 2),
        )
        assert isinstance(ft.get_permutation_order_matrix(), PermutationOrderMatrix)
        assert ft.get_permutation_order_matrix().matrix_data == [
            [(3, 1, 2), (2, 3, 1), (1, 2, 3)],
            [(1, 2, 3), (3, 1, 2), (2, 3, 1)],
            [(2, 3, 1), (1, 2, 3), (3, 1, 2)],
        ]
        ft.main_permutation_order = (3, 1, 2)
        assert isinstance(ft.get_permutation_order_matrix(), PermutationOrderMatrix)
        assert ft.get_permutation_order_matrix().matrix_data == [
            [(3, 1, 2), (2, 3, 1), (1, 2, 3)],
            [(1, 2, 3), (3, 1, 2), (2, 3, 1)],
            [(2, 3, 1), (1, 2, 3), (3, 1, 2)],
        ]
        ft.add_child(
            FractalTimelineTree(duration=TimelineDuration(10), proportions=(1, 2, 3))
        )



from unittest import TestCase

from musurgia.trees.fractaltimelinetree import PermutationIndexCalculater
from musurgia.musurgia_exceptions import PermutationIndexCalculaterNoParentIndexError


class TestFtUnit(TestCase):
    def test_get_index_error(self):
        pic = PermutationIndexCalculater(size=10)
        with self.assertRaises(PermutationIndexCalculaterNoParentIndexError):
            pic.get_index(1)
        pic = PermutationIndexCalculater(size=10, parent_index=(2, 3))
        with self.assertRaises(ValueError):
            pic.get_index(11)



from unittest import TestCase
import warnings


from musurgia.trees.fractaltimelinetree import FractalTimelineTree
from musurgia.musurgia_exceptions import (
    FractalTimelineTreePermutationIndexError,
    FractalTimelineTreeSetMainPermutationOrderFirstError,
    FractalTimelineTreeHasChildrenError,
    FractalTimelineTreeMergeWrongValuesError,
    FractalTimelineTreeHasNoChildrenError,
)
from musurgia.tests.utils_for_tests import create_test_fractal_timline_tree
from musurgia.trees.timelinetree import TimelineDuration


class TestFt(TestCase):
    def setUp(self) -> None:
        self.ft = create_test_fractal_timline_tree()

    def test_get_fractal_order(self):
        expected = """└── 0
    ├── 3
    │   ├── 2
    │   ├── 4
    │   │   ├── 2
    │   │   ├── 4
    │   │   ├── 1
    │   │   └── 3
    │   ├── 1
    │   └── 3
    │       ├── 3
    │       ├── 1
    │       ├── 4
    │       └── 2
    ├── 1
    ├── 4
    │   ├── 1
    │   ├── 2
    │   ├── 3
    │   │   ├── 2
    │   │   ├── 4
    │   │   ├── 1
    │   │   └── 3
    │   └── 4
    │       ├── 1
    │       ├── 2
    │       ├── 3
    │       └── 4
    └── 2
        ├── 4
        │   ├── 2
        │   ├── 4
        │   ├── 1
        │   └── 3
        ├── 3
        │   ├── 3
        │   ├── 1
        │   ├── 4
        │   └── 2
        ├── 2
        └── 1
"""
        self.assertEqual(
            self.ft.get_tree_representation(key=lambda node: node.get_fractal_order()),
            expected,
        )

    def test_add_wrong_child(self):
        ft = FractalTimelineTree(duration=TimelineDuration(10), proportions=(1, 2, 3))
        with self.assertRaises(TypeError):
            ft.add_child("something")

    def test_non_root_main_permutation_order(self):
        ft = FractalTimelineTree(duration=TimelineDuration(10), proportions=(1, 2, 3))
        assert ft.main_permutation_order is None
        assert self.ft.main_permutation_order == (3, 1, 4, 2)
        for node in self.ft.traverse():
            assert node.main_permutation_order == self.ft.main_permutation_order

    def test_calculate_permutation_index_error(self):
        ft = FractalTimelineTree(duration=TimelineDuration(10), proportions=(1, 2, 3))
        with self.assertRaises(FractalTimelineTreePermutationIndexError):
            ft.calculate_permutation_index()

        child = ft.add_child(
            FractalTimelineTree(duration=TimelineDuration(5), proportions=(1, 2, 3))
        )
        with self.assertRaises(FractalTimelineTreeSetMainPermutationOrderFirstError):
            child.calculate_permutation_index()
        with self.assertRaises(FractalTimelineTreeHasChildrenError):
            ft.main_permutation_order = (3, 1, 2)

    def test_generate_children_wrong_size(self):
        ft = FractalTimelineTree(
            duration=TimelineDuration(10),
            proportions=(1, 2, 3),
            main_permutation_order=(3, 1, 2),
            permutation_index=(1, 1),
        )
        with self.assertRaises(ValueError):
            ft.generate_children(4)
        with self.assertRaises(ValueError):
            ft.generate_children(-1)
        with self.assertRaises(TypeError):
            ft.generate_children("string")

    def test_get_children_fractal_orders_error(self):
        ft = FractalTimelineTree(duration=TimelineDuration(10), proportions=(1, 2, 3))
        with self.assertRaises(FractalTimelineTreeSetMainPermutationOrderFirstError):
            ft.get_children_fractal_orders()

    def test_merge_reduce_error(self):
        ft = FractalTimelineTree(
            duration=TimelineDuration(10),
            proportions=(1, 2, 3),
            main_permutation_order=(3, 1, 2),
            permutation_index=(1, 1),
        )
        with self.assertRaises(FractalTimelineTreeHasNoChildrenError):
            ft.merge_children(1, 3)
        with self.assertRaises(FractalTimelineTreeHasNoChildrenError):
            ft.reduce_children_by_condition(lambda node: node.get_value() == 10)
        ft.generate_children(3)
        with self.assertRaises(FractalTimelineTreeMergeWrongValuesError):
            ft.merge_children(1, 3)
        with self.assertRaises(ValueError):
            ft.reduce_children_by_size(2, mode="merge")
        with self.assertRaises(ValueError):
            ft.reduce_children_by_size(2, mode="merge", merge_index=20)

        with self.assertRaises(ValueError):
            ft.reduce_children_by_size(5)
        with self.assertRaises(ValueError):
            ft.reduce_children_by_size(-1)
        ft.reduce_children_by_size(0)
        assert len(ft.get_children()) == 3
        ft.reduce_children_by_size(2, mode="merge", merge_index=2)
        assert len(ft.get_children()) == 2

    def test_split_iter(self):
        ft = FractalTimelineTree(
            duration=TimelineDuration(10),
            proportions=(1, 2, 3),
            main_permutation_order=(3, 1, 2),
            permutation_index=(1, 1),
        )
        ft.split([1, 3, 1])
        assert len(ft.get_children()) == 3

    def test_split_error(self):
        with self.assertRaises(FractalTimelineTreeHasChildrenError):
            self.ft.split(1, 2, 3)

    def test_calculate_permutation_index(self):
        ft = FractalTimelineTree(duration=TimelineDuration(10), proportions=(1, 2, 3))
        with self.assertRaises(FractalTimelineTreePermutationIndexError):
            ft.calculate_permutation_index()

    def test_get_children_with_and_without_warning(self):
        ft = FractalTimelineTree(duration=TimelineDuration(10), proportions=(1, 2, 3))
        ft.add_child(
            FractalTimelineTree(duration=TimelineDuration(5), proportions=(1, 2, 3))
        )

        with warnings.catch_warnings(record=True) as w:
            ft.get_children()
            assert len(w) == 1

        with warnings.catch_warnings(record=True) as w:
            ft._get_children()
            assert len(w) == 0



from abc import abstractmethod, ABC
from typing import Optional, Protocol, Any

from musurgia.musurgia_types import ConvertibleToFloat, check_type, MarginType

__all__ = []  # type: ignore


class HasMarginsProtocol(Protocol):
    @property
    def left_margin(self) -> float: ...

    @property
    def top_margin(self) -> float: ...

    @property
    def right_margin(self) -> float: ...

    @property
    def bottom_margin(self) -> float: ...


class AbstractMargined(ABC):
    """
    An interface for setting and getting DrawObject's margin attributes.
    """

    def __init__(
        self,
        top_margin: Optional[ConvertibleToFloat] = None,
        bottom_margin: Optional[ConvertibleToFloat] = None,
        left_margin: Optional[ConvertibleToFloat] = None,
        right_margin: Optional[ConvertibleToFloat] = None,
        *args: Any,
        **kwargs: Any,
    ):
        super().__init__(*args, **kwargs)
        self._top_margin: float = 0
        self._left_margin: float = 0
        self._bottom_margin: float = 0
        self._right_margin: float = 0

        self.top_margin = top_margin  # type: ignore
        self.left_margin = left_margin  # type: ignore
        self.bottom_margin = bottom_margin  # type: ignore
        self.right_margin = right_margin  # type: ignore

    @property
    @abstractmethod
    def bottom_margin(self) -> float:
        """bottom_margin getter must be provided"""

    @bottom_margin.setter
    @abstractmethod
    def bottom_margin(self, val: Optional[ConvertibleToFloat]) -> None:
        """bottom_margin setter must be provided"""

    @property
    @abstractmethod
    def left_margin(self) -> float:
        """left_margin getter must be provided"""

    @left_margin.setter
    @abstractmethod
    def left_margin(self, val: Optional[ConvertibleToFloat]) -> None:
        """left_margin setter must be provided"""

    @property
    @abstractmethod
    def top_margin(self) -> float:
        """top_margin getter must be provided"""

    @top_margin.setter
    @abstractmethod
    def top_margin(self, val: Optional[ConvertibleToFloat]) -> None:
        """top_margin setter must be provided"""

    @property
    @abstractmethod
    def right_margin(self) -> float:
        """right_margin getter must be provided"""

    @right_margin.setter
    @abstractmethod
    def right_margin(self, val: Optional[ConvertibleToFloat]) -> None:
        """right_margin setter must be provided"""

    def get_margins(self) -> dict[MarginType, float]:
        return {
            "top": self.top_margin,
            "right": self.right_margin,
            "bottom": self.bottom_margin,
            "left": self.left_margin,
        }

    @property
    def margins(self) -> tuple[float, float, float, float]:
        return self.top_margin, self.right_margin, self.bottom_margin, self.left_margin

    @margins.setter
    def margins(self, value: tuple[float, float, float, float]) -> None:
        self.top_margin, self.right_margin, self.bottom_margin, self.left_margin = value


class Margined(AbstractMargined):
    """
    An interface for setting and getting DrawObject's margin attributes.
    """

    def __init__(
        self,
        top_margin: ConvertibleToFloat = 0,
        bottom_margin: ConvertibleToFloat = 0,
        left_margin: ConvertibleToFloat = 0,
        right_margin: ConvertibleToFloat = 0,
        *args: Any,
        **kwargs: Any,
    ):
        super().__init__(
            top_margin=top_margin,
            bottom_margin=bottom_margin,
            left_margin=left_margin,
            right_margin=right_margin,
            *args,
            **kwargs,
        )  # type: ignore

    @property
    def bottom_margin(self) -> float:
        return self._bottom_margin

    @bottom_margin.setter
    def bottom_margin(self, val: ConvertibleToFloat) -> None:
        check_type(
            val,
            "ConvertibleToFloat",
            class_name=self.__class__.__name__,
            property_name="bottom_margin",
        )
        self._bottom_margin = float(val)

    @property
    def left_margin(self) -> float:
        return self._left_margin

    @left_margin.setter
    def left_margin(self, val: ConvertibleToFloat) -> None:
        check_type(
            val,
            "ConvertibleToFloat",
            class_name=self.__class__.__name__,
            property_name="left_margin",
        )
        self._left_margin = float(val)

    @property
    def top_margin(self) -> float:
        return self._top_margin

    @top_margin.setter
    def top_margin(self, val: ConvertibleToFloat) -> None:
        check_type(
            val,
            "ConvertibleToFloat",
            class_name=self.__class__.__name__,
            property_name="top_margin",
        )
        self._top_margin = float(val)

    @property
    def right_margin(self) -> float:
        return self._right_margin

    @right_margin.setter
    def right_margin(self, val: ConvertibleToFloat) -> None:
        check_type(
            val,
            "ConvertibleToFloat",
            class_name=self.__class__.__name__,
            property_name="right_margin",
        )
        self._right_margin = float(val)



from abc import ABC
from typing import Any, Iterator

from fpdf.drawing import DeviceGray

from musurgia.musurgia_exceptions import DrawObjectInContainerHasNegativePositionError
from musurgia.musurgia_types import create_error_message
from musurgia.pdf.drawobject import DrawObject
from musurgia.pdf.labeled import Labeled
from musurgia.pdf.margined import Margined
from musurgia.pdf.pdf import Pdf
from musurgia.pdf.positioned import Positioned

__all__ = ["DrawObjectColumn", "DrawObjectRow"]


class DrawObjectContainer(DrawObject, Labeled, Positioned, Margined, ABC):
    def __init__(
        self,
        show_borders: bool = False,
        show_margins: bool = False,
        *args: Any,
        **kwargs: Any,
    ) -> None:
        super().__init__(*args, **kwargs)
        self._draw_objects: list[DrawObject] = []
        self._show_borders: bool
        self.show_borders = show_borders
        self._show_margins: bool
        self.show_margins = show_margins

    @property
    def show_borders(self) -> bool:
        return self._show_borders

    @show_borders.setter
    def show_borders(self, val: bool) -> None:
        self._show_borders = val

    @property
    def show_margins(self) -> bool:
        return self._show_margins

    @show_margins.setter
    def show_margins(self, val: bool) -> None:
        self._show_margins = val

    def draw_margins(self, pdf: Pdf) -> None:
        if self.show_margins:
            with pdf.local_context(
                draw_color=DeviceGray(0.5), dash_pattern={"dash": 1, "gap": 1}
            ):
                pdf.rect(*self.get_margin_rectangle_coordinates())

    def draw_borders(self, pdf: Pdf) -> None:
        if self.show_borders:
            with pdf.local_context(
                draw_color=DeviceGray(0.5), dash_pattern={"dash": 2, "gap": 1}
            ):
                pdf.rect(*self.get_border_rectangle_coordinates())

    def add_draw_object(self, draw_object: DrawObject) -> DrawObject:
        if not isinstance(draw_object, DrawObject):
            raise TypeError(
                create_error_message(
                    class_name=self.__class__.__name__,
                    method_name="add_draw_object",
                    argument_name="draw_object",
                    message=f"draw_object must be of type DrawObject, not {type(draw_object)}",
                )
            )
        self._draw_objects.append(draw_object)
        return draw_object

    def get_border_rectangle_coordinates(self) -> tuple[float, float, float, float]:
        return (
            self.relative_x,
            self.relative_y,
            self.get_relative_x2() - self.relative_x,
            self.get_relative_y2() - self.relative_y,
        )

    def get_margin_rectangle_coordinates(self) -> tuple[float, float, float, float]:
        return (
            self.relative_x,
            self.relative_y,
            self.get_relative_x2()
            - self.relative_x
            + self.right_margin
            + self.left_margin,
            self.get_relative_y2()
            - self.relative_y
            + self.bottom_margin
            + self.top_margin,
        )

    def get_draw_objects(self) -> list[DrawObject]:
        return self._draw_objects

    def traverse(self) -> Iterator["DrawObject"]:
        yield self
        for do in self.get_draw_objects():
            if isinstance(do, DrawObjectContainer):
                for dodo in do.traverse():
                    yield dodo
            else:
                yield do

    def _check_draw_objects_positions(self) -> None:
        for do in self.get_draw_objects():
            if do.relative_y < 0 or do.relative_x < 0:
                raise DrawObjectInContainerHasNegativePositionError()


class DrawObjectRow(DrawObjectContainer):
    def get_relative_x2(self) -> float:
        return self.relative_x + sum(
            [do.relative_x + do.get_width() for do in self.get_draw_objects()]
        )

    def get_relative_y2(self) -> float:
        return self.relative_y + max(
            [do.relative_y + do.get_height() for do in self.get_draw_objects()]
        )

    def draw(self, pdf: Pdf) -> None:
        self._check_draw_objects_positions()
        self.draw_margins(pdf)
        with pdf.pdf_draw_object_translate(self, translate_positions=False):
            self.draw_borders(pdf)
        with pdf.pdf_draw_object_translate(self):
            self.draw_above_text_labels(pdf)
            self.draw_left_text_labels(pdf)
            self.draw_below_text_labels(pdf)
            for do in self.get_draw_objects():
                do.draw(pdf)
                pdf.translate(do.get_width(), 0)


class DrawObjectColumn(DrawObjectContainer):
    def get_relative_x2(self) -> float:
        return self.relative_x + max([do.get_width() for do in self.get_draw_objects()])

    def get_relative_y2(self) -> float:
        return self.relative_y + sum(
            [do.relative_y + do.get_height() for do in self.get_draw_objects()]
        )

    def draw(self, pdf: Pdf) -> None:
        self._check_draw_objects_positions()
        self.draw_margins(pdf)
        with pdf.pdf_draw_object_translate(self, translate_positions=False):
            self.draw_borders(pdf)
        with pdf.pdf_draw_object_translate(self):
            self.draw_above_text_labels(pdf)
            self.draw_left_text_labels(pdf)
            self.draw_below_text_labels(pdf)
            for do in self.get_draw_objects():
                do.draw(pdf)
                pdf.translate(0, do.relative_y + do.get_height())



from math import ceil
from pathlib import Path

from musurgia.pdf.line import HorizontalSegmentedLine
from musurgia.pdf.ruler import HorizontalRuler
from musurgia.pdf.drawobject import ClippingArea
from musurgia.pdf.pdf import Pdf
from musurgia.pdf.pdf_tools import draw_ruler
from musurgia.pdf.rowcolumn import DrawObjectColumn
from musurgia.pdf.text import PageText
from musurgia.tests.utils_for_tests import PdfTestCase, create_test_path

path = Path(__file__)


class TestClip(PdfTestCase):
    def setUp(self) -> None:
        self.pdf = Pdf()

    def test_line(self):
        pdf_path = create_test_path(path, "line.pdf")
        self.pdf.rect(0, 0, 50, 50)
        self.pdf.clip_rect(0, 0, 50, 50)
        self.pdf.line(10, 20, 100, 100)
        self.pdf.write_to_path(pdf_path)
        self.assertCompareFiles(pdf_path)

    def test_line_break(self):
        def draw_with_clip(index):
            with self.pdf.saved_state():
                self.pdf.clip_rect(-1, -5, 196, 50)
                self.pdf.translate(index * -190, 0)
                ruler.draw(self.pdf)

        ruler = HorizontalRuler(length=800, unit=10)
        with self.file_path(path, "line_break", "pdf") as pdf_path:
            self.pdf.translate(10, 10)
            number_of_rows = int(ceil(ruler.get_length() / 190))
            for index in range(number_of_rows):
                if index != 0:
                    self.pdf.translate(0, 30)
                draw_with_clip(index)
            self.pdf.write_to_path(pdf_path)
            self.assertCompareFiles(pdf_path)

    def test_column_line_break(self):
        def draw_with_clip(index):
            with self.pdf.saved_state():
                self.pdf.clip_rect(-1, -5, 196, 50)
                self.pdf.translate(index * -190, 0)
                c.draw(self.pdf)

        c = DrawObjectColumn()
        c.bottom_margin = 10
        c.add_draw_object(HorizontalRuler(length=800))
        c.add_draw_object(HorizontalSegmentedLine(lengths=400 * [2]))
        with self.file_path(path, "column_line_break", "pdf") as pdf_path:
            self.pdf.translate_page_margins()
            number_of_rows = int(ceil(c.get_width() / 190))
            for index in range(number_of_rows):
                if index != 0:
                    self.pdf.translate(0, c.get_height())
                draw_with_clip(index)
            self.pdf.write_to_path(pdf_path)
            self.assertCompareFiles(pdf_path)

    def test_column_page_break(self):
        self.pdf.t_margin = 15

        def _prepare_page():
            self.pdf.translate_page_margins()
            draw_ruler(self.pdf, "v")
            draw_ruler(self.pdf, "h")
            self.pdf.translate(clip_area_left_margin, clip_area_top_margin)

        def _add_page():
            self.pdf.add_page()
            _prepare_page()

        def draw_with_clip(index):
            with self.pdf.saved_state():
                self.pdf.clip_rect(-1, -5, clip_area_width + 1.14, clip_area_height)
                self.pdf.translate(index * -clip_area_width, 0)
                c.draw(self.pdf)

        c = DrawObjectColumn()
        c.bottom_margin = 60
        c.add_draw_object(HorizontalRuler(length=1200, bottom_margin=5))
        c.add_draw_object(HorizontalSegmentedLine(lengths=600 * [2]))

        clip_area_left_margin = 10
        clip_area_top_margin = 10
        clip_area_width = (
            self.pdf.w - self.pdf.l_margin - self.pdf.r_margin - clip_area_left_margin
        )
        clip_area_height = c.get_height()

        with self.file_path(path, "column_page_break", "pdf") as pdf_path:
            _prepare_page()
            number_of_rows = int(ceil(c.get_width() / clip_area_width))
            for index in range(number_of_rows):
                if index != 0:
                    self.pdf.translate(0, c.get_height())
                if self.pdf.absolute_y > self.pdf.h - self.pdf.b_margin:
                    _add_page()
                draw_with_clip(index)
            self.pdf.write_to_path(pdf_path)

    def test_with_clipping_area(self):
        c = DrawObjectColumn()
        c.bottom_margin = 60
        c.add_draw_object(HorizontalRuler(length=1200, bottom_margin=5))
        c.add_draw_object(HorizontalSegmentedLine(lengths=600 * [2]))
        ca = ClippingArea(self.pdf, draw_object=c)
        ca.left_margin = 10
        self.pdf.translate_page_margins()
        ca.draw()
        with self.file_path(path, "with_clipping_area", "pdf") as pdf_path:
            self.pdf.write_to_path(pdf_path)
            self.assertCompareFiles(pdf_path)

    def test_clipped_draw(self):
        c = DrawObjectColumn()
        c.get_clipping_area().top_margin = c.get_clipping_area().left_margin = 10
        c.bottom_margin = 60
        c.add_draw_object(
            HorizontalRuler(length=1200, bottom_margin=0, show_borders=True)
        )
        c.add_draw_object(HorizontalSegmentedLine(lengths=600 * [2]))
        title = PageText(
            "A very nice title",
            h_position="center",
            font_weight="bold",
            font_size=12,
            top_margin=10,
        )
        title.draw(self.pdf)

        self.pdf.translate_page_margins()
        self.pdf.translate(0, 20)
        c.clipped_draw(self.pdf)

        for page in self.pdf.pages:
            self.pdf.page = page
            self.pdf.reset_position()
            self.pdf.translate_page_margins()
            draw_ruler(self.pdf, "h")
            draw_ruler(self.pdf, "v")
        with self.file_path(path, "clipped_draw", "pdf") as pdf_path:
            self.pdf.write_to_path(pdf_path)



from pathlib import Path
from unittest import TestCase

from musurgia.musurgia_exceptions import (
    RulerCannotSetLengthsError,
    SegmentedLineLengthsCannotBeSetError,
    RulerLengthNotPositiveError,
)
from musurgia.pdf.pdf import Pdf
from musurgia.pdf.pdf_tools import draw_ruler
from musurgia.pdf.ruler import HorizontalRuler
from musurgia.tests.utils_for_tests import PdfTestCase

path = Path(__file__)


class TestRuler(TestCase):
    def setUp(self) -> None:
        self.r = HorizontalRuler(length=10, unit=1, label_show_interval=5)

    def test_setting_length_and_lengths(self):
        with self.assertRaises(RulerCannotSetLengthsError):
            HorizontalRuler(lengths=[3, 2, 1], length=100)

        with self.assertRaises(SegmentedLineLengthsCannotBeSetError):
            self.r.lengths = [3, 2, 1]

    def test_length(self):
        hsl = HorizontalRuler(length=0)
        assert hsl.get_length() == 0
        with self.assertRaises(AttributeError):
            hsl.length = 1
        with self.assertRaises(RulerLengthNotPositiveError):
            HorizontalRuler(length=-100)

    def test_get_labels(self):
        assert [l.value for l in self.r.get_markline_text_labels()] == ["0", "5", "10"]
        assert [l.font_size for l in self.r.get_markline_text_labels()] == [10, 10, 10]
        assert [l.font_weight for l in self.r.get_markline_text_labels()] == [
            "medium",
            "medium",
            "medium",
        ]

    def test_change_labels(self):
        self.r.change_labels(font_size=6, font_weight="bold")
        assert [l.value for l in self.r.get_markline_text_labels()] == ["0", "5", "10"]
        assert [l.font_size for l in self.r.get_markline_text_labels()] == [6, 6, 6]
        assert [l.font_weight for l in self.r.get_markline_text_labels()] == [
            "bold",
            "bold",
            "bold",
        ]

        self.r.change_labels(
            condition=lambda label: True
            if self.r.segments.index(label.master.master) == 5
            else False,
            font_size=8,
        )
        assert [l.font_size for l in self.r.get_markline_text_labels()] == [6, 8, 6]

    def test_getter_setters_errors(self):
        with self.assertRaises(AttributeError):
            self.r.length
        with self.assertRaises(AttributeError):
            self.r.length = 2

        with self.assertRaises(AttributeError):
            self.r.unit
        with self.assertRaises(AttributeError):
            self.r.unit = 2

        with self.assertRaises(AttributeError):
            self.r.show_first_label
        with self.assertRaises(AttributeError):
            self.r.show_first_label = 2

        with self.assertRaises(AttributeError):
            self.r.label_show_interval
        with self.assertRaises(AttributeError):
            self.r.label_show_interval = 2

        with self.assertRaises(AttributeError):
            self.r.first_label
        with self.assertRaises(AttributeError):
            self.r.first_label = 2


class TestRulerPdf(PdfTestCase):
    def setUp(self) -> None:
        self.pdf = Pdf()

    def test_h_ruler(self):
        r = HorizontalRuler(length=50)
        with self.file_path(path, "h_ruler", "pdf") as pdf_path:
            self.pdf.translate_page_margins()
            r.draw(self.pdf)
            self.pdf.write_to_path(pdf_path)

    def test_h_ruler_A4(self):
        with self.file_path(path, "h_ruler_A4", "pdf") as pdf_path:
            self.pdf.translate_page_margins()
            draw_ruler(self.pdf, mode="h")
            self.pdf.write_to_path(pdf_path)

    def test_both_rulers_A4(self):
        with self.file_path(path, "both_rulers_A4", "pdf") as pdf_path:
            self.pdf.translate_page_margins()
            draw_ruler(self.pdf, mode="h")
            draw_ruler(self.pdf, mode="v")
            self.pdf.write_to_path(pdf_path)

    def test_rulers_borders_and_margins(self):
        with self.file_path(path, "both_rulers_borders_and_margins", "pdf") as pdf_path:
            self.pdf.translate_page_margins()
            draw_ruler(self.pdf, mode="h", show_borders=True, show_margins=True)
            draw_ruler(self.pdf, mode="v", show_borders=True, show_margins=True)
            self.pdf.write_to_path(pdf_path)



from pathlib import Path

from musurgia.musurgia_exceptions import RelativePositionNotSettableError
from musurgia.pdf.pdf import Pdf
from musurgia.pdf.text import PageText
from musurgia.tests.utils_for_tests import PdfTestCase

path = Path(__file__)


class TestPageText(PdfTestCase):
    def setUp(self):
        self.pdf = Pdf()

    def test_defaults(self):
        pt = PageText(value="nothing")
        assert pt.v_position == "top"
        assert pt.h_position == "left"
        assert pt.get_margins() == {"left": 0, "top": 0, "right": 0, "bottom": 0}
        assert pt.get_positions() == {"x": 0.0, "y": 0.0}

    def test_positions(self):
        pt = PageText(value="left top")
        pt.draw(self.pdf)
        pt = PageText(value="center top", h_position="center")
        pt.draw(self.pdf)
        pt = PageText(value="right top", h_position="right")
        pt.draw(self.pdf)

        pt = PageText(value="left bottom", v_position="bottom")
        pt.draw(self.pdf)
        pt = PageText(value="center bottom", h_position="center", v_position="bottom")
        pt.draw(self.pdf)
        pt = PageText(value="right bottom", h_position="right", v_position="bottom")
        pt.draw(self.pdf)

        with self.file_path(path, "positions", "pdf") as pdf_path:
            self.pdf.write_to_path(pdf_path)

    def test_setting_position_error(self):
        with self.assertRaises(RelativePositionNotSettableError):
            PageText(value="left top", relative_x=10)
        with self.assertRaises(RelativePositionNotSettableError):
            PageText(value="something").relative_y = 10



from unittest import TestCase

from musurgia.musurgia_exceptions import (
    RelativePositionNotSettableError,
    MarginNotSettableError,
)
from musurgia.pdf.drawobject import PositionedSlave, MarginedSlave
from musurgia.pdf.margined import Margined, AbstractMargined
from musurgia.pdf.positioned import Positioned
from musurgia.tests.utils_for_tests import DummyMaster


class Human(Positioned, Margined):
    pass


class Master(DummyMaster):
    def get_slave_position(self, slave: "PositionedSlave", position: str) -> float:
        positions = {"x": 1.0, "y": 2.0}
        return positions[position]

    def get_slave_margin(self, slave: "MarginedSlave", margin: str) -> float:
        margins = {"top": 1.0, "right": 2.0, "bottom": 3.0, "left": 4.0}
        return margins[margin]


class Slave(PositionedSlave, MarginedSlave):
    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)
        self._master = Master()

    @property
    def master(self) -> Master:
        return self._master


class WrongMaster(AbstractMargined):
    pass


class TestPositioned(TestCase):
    def setUp(self) -> None:
        self.s = Slave()

    def test_slave_positions(self):
        assert self.s.get_positions() == {"x": 1.0, "y": 2.0}

    def test_slave_set_position(self):
        with self.assertRaises(RelativePositionNotSettableError):
            Slave(relative_x=3)
        with self.assertRaises(RelativePositionNotSettableError):
            self.s.relative_x = 3
        with self.assertRaises(RelativePositionNotSettableError):
            Slave(relative_y=3)
        with self.assertRaises(RelativePositionNotSettableError):
            self.s.relative_y = 3

    def test_master_positions(self):
        m = Master()
        assert m.get_positions() == {"x": 0.0, "y": 0.0}
        m = Master(relative_x=10, relative_y=20)
        assert m.get_positions() == {"x": 10.0, "y": 20.0}
        assert m.get_slave_position(self.s, "x") == 1.0

    def test_human_positions(self):
        h = Human()
        assert h.get_positions() == {"x": 0.0, "y": 0.0}
        h = Human(relative_x=10, relative_y=20)
        assert h.get_positions() == {"x": 10.0, "y": 20.0}
        with self.assertRaises(AttributeError):
            h.get_slave_position()


class TestMargined(TestCase):
    def setUp(self) -> None:
        self.s = Slave()

    def test_slave_margins(self):
        assert self.s.get_margins() == {
            "bottom": 3.0,
            "left": 4.0,
            "right": 2.0,
            "top": 1.0,
        }

    def test_slave_set_margin(self):
        with self.assertRaises(MarginNotSettableError):
            Slave(bottom_margin=3.0)
        with self.assertRaises(MarginNotSettableError):
            self.s.right_margin = 3
        with self.assertRaises(MarginNotSettableError):
            self.s.left_margin = 3
        with self.assertRaises(MarginNotSettableError):
            self.s.top_margin = 3
        with self.assertRaises(MarginNotSettableError):
            self.s.bottom_margin = 3

    def test_master_margins(self):
        m = Master()
        assert m.get_margins() == {"bottom": 0.0, "left": 0.0, "right": 0.0, "top": 0.0}
        m = Master(left_margin=10, right_margin=20)
        assert m.get_margins() == {
            "bottom": 0.0,
            "left": 10.0,
            "right": 20.0,
            "top": 0.0,
        }
        assert m.get_slave_margin(self.s, "top") == 1.0

    def test_human_margins(self):
        h = Human()
        assert h.get_margins() == {"bottom": 0.0, "left": 0.0, "right": 0.0, "top": 0.0}
        h = Human(left_margin=10, right_margin=20)
        assert h.get_margins() == {
            "bottom": 0.0,
            "left": 10.0,
            "right": 20.0,
            "top": 0.0,
        }
        with self.assertRaises(AttributeError):
            h.get_slave_margin()

    def test_wrong_master(self):
        with self.assertRaises(TypeError):
            WrongMaster()



from fractions import Fraction
from typing import Any, Union

from musicscore.metronome import Metronome
from musicscore.quarterduration import QuarterDuration

from musurgia.chordfactory.chordfactory import AbstractChordFactory
from musurgia.timing.duration import (
    ReadonlyDuration,
    convert_duration_to_quarter_duration_value,
)
from musurgia.trees.valuedtree import ValuedTree
from musurgia.musurgia_types import ConvertibleToFraction, MusurgiaTypeError, check_type

__all__ = ["TimelineDuration", "TimelineTree"]


class TimelineDuration(ReadonlyDuration):
    def __init__(self, *args: Any, **kwargs: Any):
        super().__init__(*args, **kwargs)
        self._metronome: "Metronome" = Metronome(60)
        self._quarter_duration: "QuarterDuration" = QuarterDuration(1)

    def _set_seconds(self, seconds: ConvertibleToFraction) -> None:
        self._set_clock(hours=0, minutes=0, seconds=seconds)

    @property
    def metronome(self) -> "Metronome":
        return self._metronome

    @metronome.setter
    def metronome(self, value: "Metronome") -> None:
        if not isinstance(value, Metronome):
            raise TypeError
        self._metronome = value

    @property
    def quarter_duration(self) -> None:
        raise AttributeError("Use get_duration() instead.")

    @quarter_duration.setter
    def quarter_duration(self, value: Any) -> None:
        raise AttributeError("TimelineDuration.quarter_duration cannot be set.")

    def get_quarter_duration(self) -> QuarterDuration:
        self._quarter_duration.value = convert_duration_to_quarter_duration_value(
            self.metronome, self.calculate_in_seconds()
        )
        return self._quarter_duration

    def get_metronome(self) -> Metronome:
        return self._metronome


class TimelineTree(ValuedTree):
    def __init__(
        self,
        duration: Union[TimelineDuration, ConvertibleToFraction],
        *args: Any,
        **kwargs: Any,
    ):
        super().__init__(*args, **kwargs)
        if not isinstance(duration, TimelineDuration):
            try:
                check_type(duration, "ConvertibleToFraction")
            except MusurgiaTypeError:
                raise AttributeError(
                    "TimelineTree.update_duration: duration must be of of types TimelineDuration or ConvertibleToFraction"
                )
            duration = TimelineDuration(duration)
        self._duration: TimelineDuration = duration

    # private methods
    def _check_child_to_be_added(self, child: "TimelineTree") -> bool:
        if not isinstance(child, TimelineTree):
            raise TypeError(
                f"Wrong child type. Child must be of type FractalTree not {type(child)}"
            )
        return True

    def _set_value(self, value: ConvertibleToFraction) -> None:
        if not isinstance(value, Fraction):
            value = Fraction(value)
        self._duration._set_seconds(value)

    @property
    def duration(self) -> None:
        raise AttributeError("Use get_duration() instead.")

    def get_duration(self) -> TimelineDuration:
        return self._duration

    def get_value(self) -> Fraction:
        return self.get_duration().calculate_in_seconds()

    def update_duration(
        self, duration: Union[TimelineDuration, ConvertibleToFraction]
    ) -> None:
        if isinstance(duration, TimelineDuration):
            self._duration = duration
            new_value = self.get_value()
        elif isinstance(duration, Fraction):
            new_value = duration
        else:
            try:
                check_type(duration, "ConvertibleToFraction")
            except MusurgiaTypeError:
                raise AttributeError(
                    "TimelineTree.update_duration: duration must be of of types TimelineDuration or ConvertibleToFraction"
                )
            new_value = Fraction(duration)
        self.update_value(new_value)


class SimpleTimelineChordFactory(AbstractChordFactory):
    def __init__(
        self,
        timline_node: TimelineTree,
        show_metronome: bool = False,
        *args: Any,
        **kwargs: Any,
    ):
        super().__init__(*args, **kwargs)
        self._timeline_node: TimelineTree = timline_node
        self._show_metronome: bool
        self.show_metronome = show_metronome

    @property
    def show_metronome(self) -> bool:
        return self._show_metronome

    @show_metronome.setter
    def show_metronome(self, value: bool) -> None:
        self._show_metronome = value

    def update_chord_midis(self) -> None:
        self._chord.midis = 72

    def update_chord_quarter_duration(self) -> None:
        self._chord.quarter_duration = (
            self._timeline_node.get_duration().get_quarter_duration()
        )

    def update_chord_metronome(self) -> None:
        if self.show_metronome:
            self._chord.metronome = self._timeline_node.get_duration().get_metronome()



import os
from unittest import TestCase

from musurgia.trees.fractaltimelinetree import FractalTimelineTree
from musurgia.trees.timelinetree import TimelineDuration

path = os.path.abspath(__file__).split(".")[0]


class TestAddLayer(TestCase):
    def setUp(self):
        self.ft = FractalTimelineTree(
            duration=TimelineDuration(10),
            proportions=(1, 2, 3),
            main_permutation_order=(3, 1, 2),
            permutation_index=(1, 1),
        )

    def test_two_layers(self):
        self.ft.add_layer()
        self.ft.add_layer()
        self.assertEqual(self.ft.get_number_of_layers(), 2)
        self.assertEqual(
            [[1, 2, 3], [3, 1, 2], [2, 3, 1]],
            self.ft.get_leaves(key=lambda node: node.get_fractal_order()),
        )
        self.assertEqual(
            [[0.83, 1.67, 2.5], [0.83, 0.28, 0.56], [1.11, 1.67, 0.56]],
            self.ft.get_leaves(key=lambda node: round(float(node.get_value()), 2)),
        )

    def test_with_condition(self):
        self.ft.add_layer()
        self.ft.add_layer()
        self.ft.add_layer(
            lambda node: node.get_fractal_order() != 1,
            lambda node: node.get_value() > 0.6,
        )
        self.assertEqual(
            [[1, [1, 2, 3], [3, 1, 2]], [[3, 1, 2], 1, 2], [[1, 2, 3], [3, 1, 2], 1]],
            self.ft.get_leaves(key=lambda node: node.get_fractal_order()),
        )

    def test_child_add_layer(self):
        self.ft.add_layer()
        self.ft.get_children()[0].add_layer()
        expected = [[0.83, 1.67, 2.5], 1.67, 3.33]
        self.assertEqual(
            expected,
            self.ft.get_leaves(key=lambda leaf: round(float(leaf.get_value()), 2)),
        )

    def test_complex_layers(self):
        self.ft.add_layer()
        self.ft.add_layer(lambda n: True if n.get_fractal_order() > 1 else False)
        self.ft.add_layer(lambda n: True if n.get_fractal_order() > 1 else False)
        self.ft.add_layer(lambda n: True if n.get_fractal_order() > 1 else False)
        self.ft.add_layer(lambda n: True if n.get_fractal_order() > 1 else False)
        assert self.ft.get_layer(1, key=lambda node: node.get_fractal_order()) == [
            3,
            1,
            2,
        ]
        assert self.ft.get_layer(2, key=lambda node: node.get_fractal_order()) == [
            1,
            2,
            3,
            1,
            2,
            3,
            1,
        ]
        assert self.ft.get_layer(3, key=lambda node: node.get_fractal_order()) == [
            1,
            1,
            2,
            3,
            3,
            1,
            2,
            1,
            1,
            2,
            3,
            3,
            1,
            2,
            1,
        ]
        assert self.ft.get_layer(4, key=lambda node: node.get_fractal_order()) == [
            1,
            1,
            3,
            1,
            2,
            2,
            3,
            1,
            2,
            3,
            1,
            1,
            3,
            1,
            2,
            1,
            1,
            1,
            2,
            3,
            3,
            1,
            2,
            3,
            1,
            2,
            1,
            1,
            2,
            3,
            1,
        ]



from pathlib import Path
from unittest import TestCase

from musurgia.musurgia_exceptions import RelativePositionNotSettableError
from musurgia.pdf.line import SlaveStraightLine, StraightLine
from musurgia.pdf.pdf import Pdf
from musurgia.tests.utils_for_tests import PdfTestCase, DummyMaster

path = Path(__file__)


class TestStraightLine(TestCase):
    def setUp(self):
        self.sl_h = StraightLine(length=10, mode="h", relative_x=10, relative_y=20)
        self.sl_v = StraightLine(length=10, mode="v", relative_x=30, relative_y=40)

    def test_infos(self):
        assert self.sl_h.relative_x == 10
        assert self.sl_h.relative_y == 20
        assert self.sl_v.relative_x == 30
        assert self.sl_v.relative_y == 40

        assert self.sl_h.get_relative_x2() == 20
        assert self.sl_h.get_relative_y2() == 20
        assert self.sl_v.get_relative_x2() == 30
        assert self.sl_v.get_relative_y2() == 50

        assert self.sl_h.get_height() == 0
        assert self.sl_h.get_width() == 10
        assert self.sl_v.get_height() == 10
        assert self.sl_v.get_width() == 0

    def test_width_and_height(self):
        sl = StraightLine("h", 30)
        sl.margins = (10, 10, 10, 10)
        assert sl.get_width() == 50
        assert sl.get_height() == 20


class TestStraightLineDraw(PdfTestCase):
    def setUp(self) -> None:
        self.pdf = Pdf()

    def test_draw(self):
        h_lines = [StraightLine(length=10, mode="h") for _ in range(3)]
        v_lines = [StraightLine(length=10, mode="v") for _ in range(3)]
        with self.file_path(path, "draw", "pdf") as pdf_path:
            self.pdf.translate_page_margins()
            for l in h_lines:
                l.draw(self.pdf)
                self.pdf.translate(20, 0)
            self.pdf.reset_position()
            self.pdf.translate_page_margins()
            self.pdf.translate(10, 20)
            for l in v_lines:
                l.draw(self.pdf)
                self.pdf.translate(20, 0)
            self.pdf.write_to_path(pdf_path)


class TestSlaveStraightLine(TestCase):
    def setUp(self) -> None:
        self.master = DummyMaster()
        self.sl = SlaveStraightLine(
            mode="h", length=20, name="straight_test", master=self.master
        )

    def test_position_not_settable(self):
        with self.assertRaises(RelativePositionNotSettableError):
            SlaveStraightLine(
                relative_x=0,
                mode="h",
                length=20,
                name="straight_test",
                master=self.master,
            )
        with self.assertRaises(RelativePositionNotSettableError):
            self.sl.relative_x = 10

    def test_relative_x(self):
        actual = self.sl.relative_x
        expected = self.master.get_slave_position(self.sl, "x")
        self.assertEqual(expected, actual)

    def test_get_relative_x(self):
        actual = self.sl.get_relative_x2()
        expected = self.sl.relative_x + self.sl.length
        self.assertEqual(expected, actual)

    def test_get_width(self):
        actual = self.sl.get_width()
        expected = self.sl.left_margin + self.sl.length + self.sl.right_margin
        self.assertEqual(expected, actual)

    def test_get_height(self):
        actual = self.sl.get_height()
        expected = self.sl.top_margin + self.sl.bottom_margin
        self.assertEqual(expected, actual)

    pass


class TestSlaveStraightLineDraw(PdfTestCase):
    def setUp(self) -> None:
        self.pdf = Pdf()
        self.master = DummyMaster()
        self.sl = SlaveStraightLine(
            mode="h", length=20, name="straight_test", master=self.master
        )

    def test_draw(self):
        with self.file_path(path, "draw_slave", "pdf") as pdf_path:
            self.pdf.translate_page_margins()
            self.sl.draw(self.pdf)
            self.pdf.write_to_path(pdf_path)



from unittest import TestCase

from musurgia.matrix.matrix import (
    Matrix,
    MatrixIsEmptyError,
    SquareMatrix,
    PermutationOrderMatrix,
    PermutationOrderMatrixGenerator,
    MatrixIndexController,
)
from musurgia.musurgia_exceptions import (
    MatrixIndexOutOfRangeError,
    MatrixIndexEndOfRowError,
    MatrixIndexEndOfMatrixError,
    SquareMatrixDataError,
    PermutationOrderMatrixDataError,
    MatrixIndexControllerReadingDirectionError,
)
from musurgia.permutation.permutation import permute


class TestMatrix(TestCase):
    def setUp(self):
        self.mat = Matrix([[1, 2, 3], [4, 5, 6]])
        self.big_mat = Matrix(
            [
                [(1, 1), (1, 2), (1, 3), (1, 4)],
                [(2, 1), (2, 2), (2, 3), (2, 4)],
                [(3, 1), (3, 2), (3, 3), (3, 4)],
                [(4, 1), (4, 2), (4, 3), (4, 4)],
                [(5, 1), (5, 2), (5, 3), (5, 4)],
            ]
        )

    def test_matrix(self):
        assert Matrix(None).matrix_data == []
        m = [[1, 2, 3], [4, 5, 6]]
        mat = Matrix(m)
        assert mat.matrix_data == m
        with self.assertRaises(TypeError):
            mat.matrix_data = [1, 2, 3]

    def test_add_row(self):
        mat = Matrix()
        # wrong type
        with self.assertRaises(TypeError):
            mat.add_row(1)
        # adding two rows
        row_1 = [1, 2, 3]
        mat.add_row(row_1)
        assert mat.matrix_data == [row_1]
        row_2 = [4, 5, 6]
        mat.add_row(row_2)
        assert mat.matrix_data == [row_1, row_2]
        # wrong lengths
        with self.assertRaises(ValueError):
            mat.add_row([1, 2, 3, 4, 5])
        with self.assertRaises(ValueError):
            mat.add_row([1, 2])

    def test_remove_row(self):
        with self.assertRaises(MatrixIsEmptyError):
            Matrix().remove_row(1)
        with self.assertRaises(TypeError):
            self.mat.remove_row(0)
        with self.assertRaises(TypeError):
            self.mat.remove_row(1.2)
        with self.assertRaises(TypeError):
            self.mat.remove_row(-1)
        with self.assertRaises(ValueError):
            self.mat.remove_row(5)
        assert self.mat.remove_row(1) == [1, 2, 3]
        assert self.mat.matrix_data == [[4, 5, 6]]
        assert self.mat.get_row_size() == 3
        assert self.mat.get_column_size() == 1
        self.mat.remove_row(1)
        assert self.mat.is_empty
        assert self.mat.get_row_size() == 0
        assert self.mat.get_column_size() == 0

    def test_get_row_size(self):
        assert self.mat.get_row_size() == 3

    def test_get_column_size(self):
        assert self.mat.get_column_size() == 2

    def test_get_row(self):
        with self.assertRaises(MatrixIsEmptyError):
            Matrix().get_row(1)
        assert self.mat.get_row(1) == [1, 2, 3]
        assert self.mat.get_row(2) == [4, 5, 6]
        with self.assertRaises(TypeError):
            self.mat.get_row(0)
        with self.assertRaises(ValueError):
            self.mat.get_row(3)

    def test_get_column(self):
        with self.assertRaises(MatrixIsEmptyError):
            Matrix().get_column(1)
        assert self.mat.get_column(1) == [1, 4]
        assert self.mat.get_column(2) == [2, 5]
        assert self.mat.get_column(3) == [3, 6]
        with self.assertRaises(TypeError):
            self.mat.get_column(0)
        with self.assertRaises(ValueError):
            self.mat.get_column(4)

    def test_get_element(self):
        with self.assertRaises(MatrixIsEmptyError):
            Matrix().get_element((1, 1))
        with self.assertRaises(TypeError):
            self.mat.get_element(1)
        with self.assertRaises(ValueError):
            self.mat.get_element((3, 3))
        assert self.mat.get_element((1, 2)) == 2
        assert self.mat.get_element((2, 2)) == 5
        assert self.big_mat.get_element((1, 2)) == (1, 2)
        assert self.big_mat.get_element((3, 4)) == (3, 4)

        with self.assertRaises(MatrixIndexOutOfRangeError):
            self.mat.get_element((20, 30))

    def test_transpose(self):
        expected = [
            [(1, 1), (2, 1), (3, 1), (4, 1), (5, 1)],
            [(1, 2), (2, 2), (3, 2), (4, 2), (5, 2)],
            [(1, 3), (2, 3), (3, 3), (4, 3), (5, 3)],
            [(1, 4), (2, 4), (3, 4), (4, 4), (5, 4)],
        ]
        self.assertEqual(expected, self.big_mat.get_transposed_matrix().matrix_data)

    def test_transpose_regular(self):
        expected = [
            [(1, 1), (2, 1), (3, 1), (4, 1), (5, 1)],
            [(1, 2), (2, 2), (3, 2), (4, 2), (5, 2)],
            [(1, 3), (2, 3), (3, 3), (4, 3), (5, 3)],
            [(1, 4), (2, 4), (3, 4), (4, 4), (5, 4)],
        ]

        self.assertEqual(
            expected, self.big_mat.get_transposed_matrix(mode="regular").matrix_data
        )

    def test_transpose_diagonal(self):
        expected = [
            [(1, 1), (2, 2), (3, 3), (4, 4), (5, 1)],
            [(1, 2), (2, 3), (3, 4), (4, 1), (5, 2)],
            [(1, 3), (2, 4), (3, 1), (4, 2), (5, 3)],
            [(1, 4), (2, 1), (3, 2), (4, 3), (5, 4)],
        ]
        assert (
            self.big_mat.get_transposed_matrix(mode="diagonal").matrix_data == expected
        )

    def test_permute_matrix_data(self):
        permutation_order = (3, 1, 2, 5, 4)
        expected = [
            [(3, 1), (3, 2), (3, 3), (3, 4)],
            [(1, 1), (1, 2), (1, 3), (1, 4)],
            [(2, 1), (2, 2), (2, 3), (2, 4)],
            [(5, 1), (5, 2), (5, 3), (5, 4)],
            [(4, 1), (4, 2), (4, 3), (4, 4)],
        ]
        assert (
            permute(
                input_list=self.big_mat.matrix_data, permutation_order=permutation_order
            )
            == expected
        )

    #
    # def test_permute_columns(self):
    #     assert False


class TestSquareMatrix(TestCase):
    def test_init_errors(self):
        with self.assertRaises(TypeError):
            SquareMatrix()
        with self.assertRaises(SquareMatrixDataError):
            SquareMatrix(matrix_data=[])
        with self.assertRaises(SquareMatrixDataError):
            SquareMatrix(matrix_data=[[1, 2, 3]])

    def test_get_size(self):
        assert (
            SquareMatrix(matrix_data=[[1, 2, 3], [4, 5, 6], [7, 8, 9]]).get_size() == 3
        )


class TestGeneratePermutationOrderMatrix(TestCase):
    def test_wrong_size(self):
        with self.assertRaises(TypeError):
            PermutationOrderMatrixGenerator(main_permutation_order=[1, 2])

    def test_generation(self):
        assert PermutationOrderMatrixGenerator(
            main_permutation_order=(2, 1)
        ).generate_permutation_order_matrix().matrix_data == [
            [(2, 1), (1, 2)],
            [(1, 2), (2, 1)],
        ]
        assert PermutationOrderMatrixGenerator(
            main_permutation_order=(3, 1, 2)
        ).generate_permutation_order_matrix().matrix_data == [
            [(3, 1, 2), (2, 3, 1), (1, 2, 3)],
            [(1, 2, 3), (3, 1, 2), (2, 3, 1)],
            [(2, 3, 1), (1, 2, 3), (3, 1, 2)],
        ]
        assert PermutationOrderMatrixGenerator(
            main_permutation_order=(3, 1, 4, 2)
        ).generate_permutation_order_matrix().matrix_data == [
            [(3, 1, 4, 2), (4, 3, 2, 1), (2, 4, 1, 3), (1, 2, 3, 4)],
            [(2, 4, 1, 3), (3, 1, 4, 2), (1, 2, 3, 4), (4, 3, 2, 1)],
            [(1, 2, 3, 4), (2, 4, 1, 3), (4, 3, 2, 1), (3, 1, 4, 2)],
            [(4, 3, 2, 1), (1, 2, 3, 4), (3, 1, 4, 2), (2, 4, 1, 3)],
        ]
        pomg = PermutationOrderMatrixGenerator(main_permutation_order=(3, 1, 4, 2))
        assert pomg.main_permutation_order == (3, 1, 4, 2)
        pomg.main_permutation_order = (1, 2)
        assert pomg.main_permutation_order == (1, 2)


class TestPermutationOrderMatrix(TestCase):
    def setUp(self):
        self.pom = PermutationOrderMatrixGenerator(
            main_permutation_order=(3, 1, 2)
        ).generate_permutation_order_matrix()
        """
        [[(3, 1, 2), (2, 3, 1), (1, 2, 3)],
         [(1, 2, 3), (3, 1, 2), (2, 3, 1)],
         [(2, 3, 1), (1, 2, 3), (3, 1, 2)]]
        """

    def test_init_errors(self):
        with self.assertRaises(TypeError):
            PermutationOrderMatrix()
        with self.assertRaises(SquareMatrixDataError):
            PermutationOrderMatrix(matrix_data=[])
        # wrong type
        with self.assertRaises(PermutationOrderMatrixDataError):
            PermutationOrderMatrix(matrix_data=[[1, 2, 3]])
        # wrong permutation order
        with self.assertRaises(PermutationOrderMatrixDataError):
            PermutationOrderMatrix(matrix_data=[[(1, 2, 4)]])
        with self.assertRaises(PermutationOrderMatrixDataError):
            PermutationOrderMatrix(matrix_data=[[(2, 4, 3)]])
        # wrong lengths
        with self.assertRaises(PermutationOrderMatrixDataError):
            PermutationOrderMatrix(
                matrix_data=[[(1, 2, 3), (1, 2, 3, 4)], [(3, 2, 1), (3, 2, 1)]]
            )
        PermutationOrderMatrix(matrix_data=[[(1, 2), (1, 2)], [(2, 1), (1, 2)]])


class TestMatrixIndexController(TestCase):
    def test_get_next(self):
        with self.assertRaises(MatrixIndexOutOfRangeError):
            MatrixIndexController(
                number_of_columns=3, number_of_rows=5, first_index=(4, 7)
            )
        controller = MatrixIndexController(number_of_columns=3, number_of_rows=5)
        controller.first_index = (1, 1)
        assert next(controller) == (1, 1)
        assert next(controller) == (1, 2)
        controller.reset()
        assert controller.first_index == (1, 1)
        controller.first_index = (1, 3)
        next(controller)
        assert next(controller) == (2, 1)
        controller.first_index = (3, 2)
        assert next(controller) == (3, 2)
        assert controller.get_next_in_row() == (3, 3)
        with self.assertRaises(MatrixIndexEndOfRowError):
            controller.get_next_in_row()
        assert next(controller) == (4, 1)
        with self.assertRaises(MatrixIndexOutOfRangeError):
            controller.first_index = (7, 2)
        controller.first_index = (5, 3)
        next(controller)
        with self.assertRaises(MatrixIndexEndOfRowError):
            next(controller)
        with self.assertRaises(MatrixIndexEndOfMatrixError):
            next(controller)
        controller.first_index = (1, 2)
        next(controller)
        controller.reset()
        assert next(controller) == (1, 2)

    def test_iterate(self):
        controller = MatrixIndexController(3, 4)
        assert [_ for _ in controller] == [
            (1, 1),
            (1, 2),
            (1, 3),
            (1, 4),
            (2, 1),
            (2, 2),
            (2, 3),
            (2, 4),
            (3, 1),
            (3, 2),
            (3, 3),
            (3, 4),
        ]

    def test_convert_flatten_index_to_index_horizontal(self):
        controller = MatrixIndexController(3, 4)
        indices = [
            controller._convert_flatten_index_to_index(flatten_index)
            for flatten_index in range(0, 12)
        ]
        assert indices == [
            (1, 1),
            (1, 2),
            (1, 3),
            (1, 4),
            (2, 1),
            (2, 2),
            (2, 3),
            (2, 4),
            (3, 1),
            (3, 2),
            (3, 3),
            (3, 4),
        ]
        with self.assertRaises(MatrixIndexEndOfMatrixError):
            controller._convert_flatten_index_to_index(12)

    def test_convert_flatten_index_to_index_diagonal(self):
        controller = MatrixIndexController(3, 4, reading_direction="diagonal")
        indices = [
            controller._convert_flatten_index_to_index(flatten_index)
            for flatten_index in range(0, 12)
        ]
        assert indices == [
            (1, 1),
            (2, 2),
            (3, 3),
            (1, 2),
            (2, 3),
            (3, 4),
            (1, 3),
            (2, 4),
            (3, 1),
            (1, 4),
            (2, 1),
            (3, 2),
        ]
        with self.assertRaises(MatrixIndexEndOfMatrixError):
            controller._convert_flatten_index_to_index(12)

    def test_convert_flatten_index_to_index_vertical(self):
        controller = MatrixIndexController(3, 4, reading_direction="vertical")
        indices = [
            controller._convert_flatten_index_to_index(flatten_index)
            for flatten_index in range(0, 12)
        ]
        assert indices == [
            (1, 1),
            (2, 1),
            (3, 1),
            (1, 2),
            (2, 2),
            (3, 2),
            (1, 3),
            (2, 3),
            (3, 3),
            (1, 4),
            (2, 4),
            (3, 4),
        ]
        with self.assertRaises(MatrixIndexEndOfMatrixError):
            controller._convert_flatten_index_to_index(12)

    def test_convert_index_to_flatten_index(self):
        controller = MatrixIndexController(3, 4)
        for flatten_index in range(0, 12):
            assert (
                controller._convert_index_to_flatten_index(
                    controller._convert_flatten_index_to_index(flatten_index)
                )
                == flatten_index
            )

        controller.reading_direction = "diagonal"
        for flatten_index in range(0, 12):
            assert (
                controller._convert_index_to_flatten_index(
                    controller._convert_flatten_index_to_index(flatten_index)
                )
                == flatten_index
            )

        controller.reading_direction = "vertical"
        for flatten_index in range(0, 12):
            assert (
                controller._convert_index_to_flatten_index(
                    controller._convert_flatten_index_to_index(flatten_index)
                )
                == flatten_index
            )

    def test_get_next_flatten_index(self):
        controller = MatrixIndexController(3, 4)
        for x in range(0, 12):
            assert controller.get_next_flatten_index() == x
            next(controller)
        controller.reading_direction = "diagonal"
        for x in range(0, 12):
            assert controller.get_next_flatten_index() == x
            next(controller)

        controller = MatrixIndexController(3, 3)
        for x in range(0, 9):
            assert controller.get_next_flatten_index() == x
            next(controller)
        controller.reading_direction = "diagonal"
        for x in range(0, 9):
            assert controller.get_next_flatten_index() == x
            next(controller)

    def test_reading_direction(self):
        controller = MatrixIndexController(2, 3)
        assert controller.reading_direction == "horizontal"
        controller.reading_direction = "diagonal"
        with self.assertRaises(MatrixIndexControllerReadingDirectionError):
            controller.get_next_in_row()
        controller.first_index = (1, 2)
        assert next(controller) == (1, 2)
        assert next(controller) == (2, 3)
        assert next(controller) == (1, 3)
        assert next(controller) == (2, 1)
        with self.assertRaises(MatrixIndexEndOfMatrixError):
            next(controller)
        controller = MatrixIndexController(
            3, 3, first_index=(1, 3), reading_direction="diagonal"
        )
        assert next(controller) == (1, 3)
        assert next(controller) == (2, 1)
        assert next(controller) == (3, 2)
        with self.assertRaises(MatrixIndexEndOfMatrixError):
            next(controller)
        """
        [[(1, 1), (1, 2), (1, 3), (1, 4)],
         [(2, 1), (2, 2), (2, 3), (2, 4)],
         [(3, 1), (3, 2), (3, 3), (3, 4)],
         [(4, 1), (4, 2), (4, 3), (4, 4)],
         [(5, 1), (5, 2), (5, 3), (5, 4)]]
        """
        controller = MatrixIndexController(5, 4, reading_direction="diagonal")
        expected = [
            (1, 1),
            (2, 2),
            (3, 3),
            (4, 4),
            (5, 1),
            (1, 2),
            (2, 3),
            (3, 4),
            (4, 1),
            (5, 2),
            (1, 3),
            (2, 4),
            (3, 1),
            (4, 2),
            (5, 3),
            (1, 4),
            (2, 1),
            (3, 2),
            (4, 3),
            (5, 4),
        ]
        assert list(controller) == expected



import copy
from pathlib import Path
from unittest import TestCase

from musurgia.trees.graphictree import (
    GraphicTree,
    ValuedTreeNodeSegment,
    GraphicChildrenSegmentedLine,
)
from musurgia.musurgia_exceptions import SegmentedLineSegmentHasMarginsError
from musurgia.pdf import DrawObjectColumn, DrawObjectRow, Pdf, draw_ruler, TextLabel
from musurgia.pdf.ruler import HorizontalRuler
from musurgia.tests.utils_for_tests import (
    PdfTestCase,
    create_test_valued_tree,
    add_node_infos_to_graphic,
)

path = Path(__file__)


class GraphicTreeTest(TestCase):
    def setUp(self):
        self.vt = create_test_valued_tree()
        # print(self.vt.get_tree_representation(key=lambda node: f"{node.get_position_in_tree()}: {node.get_duration().seconds}"))
        self.gt = GraphicTree(self.vt)

    def test_populate(self):
        traversed_ft = list(self.vt.traverse())
        traversed_gt = list(self.gt.traverse())
        assert len(traversed_ft) == len(traversed_gt)
        for fnode, gnode in zip(traversed_ft, traversed_gt):
            assert gnode.get_segment().length == float(fnode.get_value() * self.gt.unit)

    def test_set_node_value(self):
        seg = ValuedTreeNodeSegment(node_value=10)
        assert seg.length == 10
        assert seg.get_node_value() == 10
        seg.set_node_value(20)
        assert seg.get_node_value() == 20
        assert seg.length == 20
        seg.unit = 2
        assert seg.get_node_value() == 20
        assert seg.length == 40
        seg.set_node_value(10)
        assert seg.get_node_value() == 10
        assert seg.length == 20

    def test_node_segment(self):
        assert self.gt.get_segment().unit == 1
        segment = self.gt.get_segment()
        assert isinstance(segment, ValuedTreeNodeSegment)
        assert segment.length == self.vt.get_value()
        with self.assertRaises(AttributeError):
            segment.length = 10
        assert segment.start_mark_line.length == segment.end_mark_line.length == 6
        segment.unit = 10
        assert self.gt.get_segment().unit == 10
        assert segment.length == self.vt.get_value() * 10
        assert segment.start_mark_line.show
        assert not segment.end_mark_line.show
        segment.end_mark_line.show = True

        segment.start_mark_line.add_text_label("something")
        copied_segment = copy.deepcopy(segment)
        assert isinstance(copied_segment, ValuedTreeNodeSegment)
        assert copied_segment.unit == segment.unit
        assert copied_segment.length == self.vt.get_value() * 10
        assert copied_segment.start_mark_line.show
        assert segment.end_mark_line.show
        copied_segment.unit = 5
        assert segment.unit == 10
        copied_segment.end_mark_line.show = False
        assert segment.end_mark_line.show
        assert not copied_segment.end_mark_line.show
        assert copied_segment.start_mark_line.get_text_labels()[0].value == "something"

    def test_change_unit(self):
        self.gt.unit = 10
        for node in self.gt.traverse():
            assert node.unit == 10

    def test_graphic(self):
        for node in self.gt.traverse():
            assert isinstance(node.get_graphic(), DrawObjectColumn)
            if node.is_leaf:
                number_draw_objects = 1
            else:
                number_draw_objects = 2
            assert len(node.get_graphic().get_draw_objects()) == number_draw_objects
            assert node.get_graphic().get_draw_objects()[0] == node.get_segment()
            if not node.is_leaf:
                assert isinstance(
                    node.get_graphic().get_draw_objects()[1], DrawObjectRow
                )
                for ch, do in zip(
                    node.get_children(),
                    node.get_graphic().get_draw_objects()[1].get_draw_objects(),
                ):
                    assert isinstance(do, DrawObjectColumn)
                    assert do.get_draw_objects()[0] == ch.get_segment()

    def test_end_mark_lines(self):
        self.gt._show_last_mark_lines()
        for node in self.gt.traverse():
            if node.is_last_child:
                assert node.get_segment().end_mark_line.show is True
            else:
                assert node.get_segment().end_mark_line.show is False

    def test_mark_line_lengths(self):
        for node in self.gt.traverse():
            node._update_mark_line_length()
        for node in self.gt.traverse():
            if node.is_first_child:
                assert (
                    node.get_segment().start_mark_line.length == node.mark_line_length
                )
            else:
                self.assertAlmostEqual(
                    node.mark_line_length * node.shrink_factor,
                    node.get_segment().start_mark_line.length,
                )
            if node.is_last_child:
                assert node.get_segment().end_mark_line.length == node.mark_line_length

    def test_set_all_distances(self):
        self.gt.set_all_distances(20)
        for node in self.gt.traverse():
            assert node.distance == 20

    def test_wrong_child_type(self):
        with self.assertRaises(TypeError):
            self.gt.add_child("string")


class TestGraphicTreeDraw(PdfTestCase):
    def setUp(self):
        self.vt = create_test_valued_tree()
        self.gt = GraphicTree(self.vt)
        self.pdf = Pdf(orientation="l")
        add_node_infos_to_graphic(self.vt, self.gt)

    def test_draw_graphic(self):
        unit = 25
        self.gt.unit = unit
        self.gt.set_all_distances(20)

        with self.file_path(path, "draw", "pdf") as pdf_path:
            self.pdf.translate_page_margins()
            draw_ruler(self.pdf, "h", unit=10, first_label=-1)
            draw_ruler(self.pdf, "v", unit=10)
            self.pdf.translate(10, 10)
            self.gt.draw(self.pdf)
            self.pdf.write_to_path(pdf_path)

    def test_draw_last_child_graphic(self):
        child = self.vt.get_children()[-1]
        child.update_value(15)
        graphic = GraphicTree(child, unit=10, distance=20)

        for index, (gn, fn) in enumerate(zip(graphic.traverse(), child.traverse())):
            text_label = TextLabel(f"{fn.get_position_in_tree()}", font_size=8)
            if gn.get_distance() > 1 and index % 2 == 1:
                text_label.placement = "below"
                text_label.top_margin = 1
            else:
                text_label.bottom_margin = 2
            gn.get_segment().start_mark_line.add_text_label(text_label)

        with self.file_path(path, "draw_last_child", "pdf") as pdf_path:
            self.pdf.translate_page_margins()
            draw_ruler(self.pdf, "h", first_label=-1)
            draw_ruler(self.pdf, "v")
            self.pdf.translate(10, 10)
            graphic.draw(self.pdf)
            self.pdf.write_to_path(pdf_path)

    def test_draw_clipped(self):
        unit = 15
        self.vt.update_value(50)
        gt = GraphicTree(self.vt, unit=unit, distance=5)
        add_node_infos_to_graphic(self.vt, gt)
        gt.set_all_distances(10)
        graphic = gt.get_graphic()
        graphic.bottom_margin = 15
        graphic._draw_objects = graphic._draw_objects[1:]
        c = DrawObjectColumn()
        c.bottom_margin = 10
        ruler = HorizontalRuler(unit=unit, length=graphic.get_width())
        c.add_draw_object(ruler)
        c.add_draw_object(graphic)

        with self.file_path(path, "draw_clipped", "pdf") as pdf_path:
            self.pdf.r_margin = self.pdf.l_margin = (self.pdf.w - 18 * unit) / 2
            self.pdf.translate_page_margins()
            c.clipped_draw(self.pdf)
            self.pdf.write_to_path(pdf_path)


class TestGraphicChildrenSegmentedLine(TestCase):
    def setUp(self):
        self.vt = create_test_valued_tree()
        self.gt = GraphicTree(self.vt)

    def test_check_segment_margins(self):
        gcsl = GraphicChildrenSegmentedLine(self.gt.get_children())
        gcsl.segments[0].margins = (1, 2, 3, 4)
        with self.assertRaises(SegmentedLineSegmentHasMarginsError):
            gcsl.draw(Pdf())

    def test_set_straight_line_relative_y(self):
        gcsl = GraphicChildrenSegmentedLine(self.gt.get_children())
        assert set([seg.get_height() for seg in gcsl.segments]) == {4.2}
        gcsl.set_straight_line_relative_y(0)
        assert set([seg.relative_y for seg in gcsl.segments]) == {0}
        assert gcsl.relative_y == -2.1


class TestCreatLayerGraphic(PdfTestCase):
    def setUp(self):
        self.vt = create_test_valued_tree()
        self.pdf = Pdf(orientation="l")

    def test_create_layer_graphic(self):
        unit = 24
        gt = GraphicTree(self.vt, unit=unit, distance=12, shrink_factor=0.6)
        add_node_infos_to_graphic(self.vt, gt)

        graphic = gt.get_graphic()
        graphic.bottom_margin = 30

        graphic_layer_2 = gt.create_layer_graphic(layer_number=2)
        graphic_layer_2.add_text_label(
            TextLabel(value="layer 2", placement="left", font_size=8, right_margin=2)
        )
        graphic_layer_2.bottom_margin = 20

        graphic_layer_3 = gt.create_layer_graphic(layer_number=3)
        graphic_layer_3.add_text_label(
            TextLabel(value="layer 3", placement="left", font_size=8, right_margin=2)
        )

        c = DrawObjectColumn()

        c.add_draw_object(graphic)
        c.add_draw_object(graphic_layer_2)
        c.add_draw_object(graphic_layer_3)

        with self.file_path(path, "draw_layers", "pdf") as pdf_path:
            self.pdf.translate_page_margins()
            draw_ruler(self.pdf, "h", unit=unit, first_label=-1)
            draw_ruler(self.pdf, "v", unit=unit)
            self.pdf.translate(unit, 10)
            c.draw(self.pdf)
            self.pdf.write_to_path(pdf_path)



from fractions import Fraction
from typing import Any, Optional

from musurgia.musurgia_exceptions import (
    ClockWrongSecondsTypeError,
    ClockWrongSecondsValueError,
    ClockWrongMinutesTypeError,
    ClockWrongHoursTypeError,
    ClockWrongMinutesValueError,
)
from musurgia.musurgia_types import (
    check_type,
    create_error_message,
    ClockMode,
    ConvertibleToFraction,
)


class Clock:
    def __init__(
        self,
        hours: int = 0,
        minutes: int = 0,
        seconds: Fraction = Fraction(0),
        *args: Any,
        **kwargs: Any,
    ) -> None:
        super().__init__(*args, **kwargs)
        self._hours: int
        self._minutes: int
        self._seconds: Fraction

        self.hours = hours
        self.minutes = minutes
        self.seconds = seconds

    @property
    def hours(self) -> int:
        return self._hours

    @hours.setter
    def hours(self, val: int) -> None:
        try:
            check_type(
                val, int, class_name=self.__class__.__name__, property_name="hours"
            )
        except TypeError as err:
            raise ClockWrongHoursTypeError(err)
        self._hours = val

    @property
    def minutes(self) -> int:
        return self._minutes

    @minutes.setter
    def minutes(self, val: int) -> None:
        try:
            check_type(
                val, int, class_name=self.__class__.__name__, property_name="minutes"
            )
        except TypeError as err:
            raise ClockWrongMinutesTypeError(err)
        if -60 <= val >= 60:
            msg = create_error_message(
                message=f"{val} must be between -60 and 60",
                class_name=self.__class__.__name__,
                property_name="minutes",
            )
            raise ClockWrongMinutesValueError(msg)
        self._minutes = val

    @property
    def seconds(self) -> Fraction:
        return self._seconds

    @seconds.setter
    def seconds(self, val: ConvertibleToFraction) -> None:
        try:
            check_type(
                val,
                "ConvertibleToFraction",
                class_name=self.__class__.__name__,
                property_name="seconds",
            )
        except TypeError as err:
            raise ClockWrongSecondsTypeError(err)
        if -60 <= val >= 60:
            msg = create_error_message(
                message=f"{val} must be between -60 and 60",
                class_name=self.__class__.__name__,
                property_name="seconds",
            )
            raise ClockWrongSecondsValueError(msg)
        self._seconds = Fraction(val)

    def get_values(self) -> tuple[int, int, Fraction]:
        return self.hours, self.minutes, self.seconds

    def set_values(
        self, hours: int, minutes: int, seconds: ConvertibleToFraction
    ) -> None:
        self.hours, self.minutes, self.seconds = hours, minutes, Fraction(seconds)

    def get_as_string(
        self, mode: ClockMode = "hms", round_: Optional[int] = None
    ) -> str:
        check_type(
            mode,
            "ClockMode",
            class_name=self.__class__.__name__,
            method_name="get_as_string",
            argument_name="mode",
        )
        s, m, h = float(self.seconds), self.minutes, self.hours
        if round_:
            s = round(s, round_)

        if m // 10 == 0 and mode != "msreduced":
            string_m = "0" + str(m)
        else:
            string_m = str(m)

        if int(s // 10) == 0 and mode != "msreduced":
            string_s = "0" + str(s)
        else:
            string_s = str(s)

        string_h = str(h)

        if not mode or mode == "hms":
            return string_h + ":" + string_m + ":" + string_s

        elif mode == "ms":
            return string_m + ":" + string_s
        elif mode == "msreduced":
            if string_m == "0":
                return string_s
            else:
                return string_m + ":" + string_s

    def calculate_in_seconds(self) -> Fraction:
        return self.convert_clock_to_seconds(self.hours, self.minutes, self.seconds)

    @staticmethod
    def convert_clock_to_seconds(
        hours: ConvertibleToFraction,
        minutes: ConvertibleToFraction,
        seconds: ConvertibleToFraction,
    ) -> Fraction:
        check_type(
            hours,
            "ConvertibleToFloat",
            class_name="Clock",
            method_name="convert_clock_to_seconds",
            argument_name="hours",
        )
        check_type(
            minutes,
            "ConvertibleToFloat",
            class_name="Clock",
            method_name="convert_clock_to_seconds",
            argument_name="minutes",
        )
        check_type(
            seconds,
            "ConvertibleToFloat",
            class_name="Clock",
            method_name="convert_clock_to_seconds",
            argument_name="seconds",
        )
        return Fraction(seconds) + (60 * Fraction(minutes)) + (3600 * Fraction(hours))

    @staticmethod
    def convert_seconds_to_clock(seconds: ConvertibleToFraction) -> "Clock":
        seconds = Fraction(seconds)
        h = int(seconds / 3600)
        s = seconds - h * 3600
        m = int(s / 60)
        s = s - m * 60

        return Clock(hours=h, minutes=m, seconds=s)

    def add_clock(self, clock: "Clock") -> "Clock":
        seconds = self.calculate_in_seconds() + clock.calculate_in_seconds()
        return self.convert_seconds_to_clock(seconds)

    def __add__(self, other: "Clock") -> "Clock":
        return self.add_clock(other)

    def subtract_clock(self, clock: "Clock") -> "Clock":
        seconds = self.calculate_in_seconds() - clock.calculate_in_seconds()
        return self.convert_seconds_to_clock(seconds)

    def __sub__(self, other: "Clock") -> "Clock":
        return self.subtract_clock(other)



from abc import ABC, ABCMeta, abstractmethod
from typing import Any

from musicscore.chord import Chord


class ChordFactoryType(ABCMeta):
    def __new__(mcls, name, bases, namespace, /, **kwargs):  # type: ignore
        chord_update_methods = {
            k for k in namespace.keys() if k.startswith("update_chord_")
        }

        for base in bases:
            base_methods = getattr(base, "_CHORD_UPDATE_METHODS", set())  # type: ignore
            chord_update_methods.update(base_methods)

        namespace["_CHORD_UPDATE_METHODS"] = chord_update_methods
        return super().__new__(mcls, name, bases, namespace, **kwargs)


class AbstractChordFactory(ABC, metaclass=ChordFactoryType):
    def __init__(self, *args: Any, **kwargs: Any):
        super().__init__(*args, **kwargs)
        # self._chord: Chord = Chord(60, 1)

    def _update_chord(self) -> None:
        for method_name in self._CHORD_UPDATE_METHODS:  # type: ignore
            getattr(self, method_name)()

    @property
    def chord(self) -> None:
        raise AttributeError("Use get_chord() instead.")

    @chord.setter
    def chord(self, value: Any) -> None:
        raise AttributeError("ChordFactory.chord cannot be set. ")

    @abstractmethod
    def update_chord_quarter_duration(self) -> None:
        pass

    @abstractmethod
    def update_chord_midis(self) -> None:
        pass

    def create_chord(self) -> "Chord":
        self._chord = Chord(60, 1)
        self._update_chord()
        return self._chord



from pathlib import Path

from musurgia.pdf.line import VerticalLineSegment
from musurgia.pdf.pdf import Pdf
from musurgia.pdf.pdf_tools import draw_ruler
from musurgia.tests.utils_for_tests import PdfTestCase

path = Path(__file__)


class TestVLSSStraighLineLabels(PdfTestCase):
    def setUp(self) -> None:
        self.pdf = Pdf()
        self.vls = VerticalLineSegment(length=20)
        self.vls.start_mark_line.show = False
        self.vls.set_straight_line_relative_x(0)

    def test_vertical_above(self):
        self.vls.straight_line.add_text_label("one above")
        self.vls.straight_line.add_text_label("two above")
        self.vls.straight_line.add_text_label("three above")
        with self.file_path(path, "above", "pdf") as pdf_path:
            self.pdf.translate_page_margins()
            draw_ruler(self.pdf, "h")
            draw_ruler(self.pdf, "v")
            self.pdf.translate(20, 20)
            self.vls.draw(self.pdf)
            self.pdf.write_to_path(pdf_path)

    def test_vertical_above_with_relative_y(self):
        self.vls.relative_y = -10
        self.vls.straight_line.add_text_label("one above")
        self.vls.straight_line.add_text_label("two above")
        self.vls.straight_line.add_text_label("three above")
        with self.file_path(path, "above_with_relative_y", "pdf") as pdf_path:
            self.pdf.translate_page_margins()
            draw_ruler(self.pdf, "h")
            draw_ruler(self.pdf, "v")
            self.pdf.translate(20, 20)
            self.vls.draw(self.pdf)
            self.pdf.write_to_path(pdf_path)

    def test_vertical_below(self):
        self.vls.straight_line.add_text_label("one below", placement="below")
        self.vls.straight_line.add_text_label("two below", placement="below")
        self.vls.straight_line.add_text_label("three below", placement="below")
        with self.file_path(path, "below", "pdf") as pdf_path:
            self.pdf.translate_page_margins()
            draw_ruler(self.pdf, "h")
            draw_ruler(self.pdf, "v")
            self.pdf.translate(20, 20)
            self.vls.draw(self.pdf)
            self.pdf.write_to_path(pdf_path)

    def test_vertical_left(self):
        self.vls.straight_line.add_text_label("one left", placement="left")
        self.vls.straight_line.add_text_label("two left", placement="left")
        self.vls.straight_line.add_text_label("three left", placement="left")
        with self.file_path(path, "left", "pdf") as pdf_path:
            self.pdf.translate_page_margins()
            draw_ruler(self.pdf, "h")
            draw_ruler(self.pdf, "v")
            self.pdf.translate(20, 20)
            self.vls.draw(self.pdf)
            self.pdf.write_to_path(pdf_path)

    def test_vertical_left_with_top_margin(self):
        self.vls.top_margin = 200
        self.vls.straight_line.add_text_label("one left", placement="left")
        self.vls.straight_line.add_text_label("two left", placement="left")
        self.vls.straight_line.add_text_label("three left", placement="left")
        with self.file_path(path, "left_with_top_margin", "pdf") as pdf_path:
            self.pdf.translate_page_margins()
            draw_ruler(self.pdf, "h")
            draw_ruler(self.pdf, "v")
            self.pdf.translate(20, 20)
            self.vls.draw(self.pdf)
            self.pdf.write_to_path(pdf_path)



from typing import Optional, Any, Iterator, Union, TypeVar

from musurgia.musurgia_types import NonNegativeInteger, check_type

__all__ = ["MagicRandom"]

T = TypeVar("T", bound="MagicRandom")


class MagicRandom:
    """
    .. code-block:: python

        from musurgia.random import MagicRandom

    MagicRandom is a class for creating pseudo random series of values. Values are chosen from a list of values
    called a 'pool' which does not contain any duplicates. The property 'periodicity' defines the minimum number of
    other values which must be given out before a value can appear again.

    >>> first = MagicRandom(pool=[1, 3, 2, 4, 5], periodicity=2, seed=20)
    >>> second = MagicRandom(pool=[1, 3, 2, 4, 5], periodicity=2, seed=20)
    >>> [first.__next__() for _ in range(20)]
    [3, 2, 1, 5, 3, 1, 4, 3, 2, 4, 5, 3, 2, 4, 1, 5, 4, 1, 3, 5]
    >>> [second.__next__() for _ in range(20)]
    [3, 1, 2, 3, 5, 1, 2, 3, 5, 1, 2, 3, 5, 4, 3, 2, 5, 4, 3, 1]
    >>> second.periodicity = 1
    >>> [second.__next__() for _ in range(20)]
    [3, 5, 4, 2, 3, 1, 5, 1, 5, 3, 4, 3, 1, 2, 1, 2, 4, 5, 1, 3]
    >>> second.periodicity = 0
    >>> [second.__next__() for _ in range(20)]
    [1, 4, 5, 2, 1, 1, 5, 1, 1, 2, 2, 1, 2, 1, 1, 3, 3, 5, 5, 1]

    """

    import random

    current_random = random

    def __init__(
        self,
        pool: list[Any],
        periodicity: Optional[NonNegativeInteger] = None,
        forbidden_list: Optional[list[Any]] = None,
        seed: Optional[Union[int, str, bytes, bytearray]] = None,
        *args: Any,
        **kwargs: Any,
    ):
        super().__init__(*args, **kwargs)
        self._pool: list[Any] = []
        self._periodicity: Optional[NonNegativeInteger] = None
        self._forbidden_list: list[Any] = []
        self._seed: Optional[Union[int, str, bytes, bytearray]] = None
        self._counter = 0
        self._previous_elements: list[Any] = []

        self.pool = pool
        self.periodicity = periodicity  # type: ignore
        self.forbidden_list = forbidden_list if forbidden_list is not None else []
        self.seed = seed

    # properties
    @property
    def counter(self) -> NonNegativeInteger:
        """
        Keeps track of the number of times :obj:`permutation_order_iterator` is called

        :return: NonNegativeInteger

        >>> r = MagicRandom(pool=['a', 'b', 'c'], periodicity=1, seed=20)
        >>> r.counter
        0
        >>> next(r)
        'c'
        >>> r.counter
        1
        >>> next(r)
        'a'
        >>> r.counter
        2
        """
        return self._counter

    @property
    def forbidden_list(self) -> list[Any]:
        """
        Set and get ``forbidden_list`` property which is used internally to keep trace of previous elements and has
        maximum length of :obj:`periodicity`. All elements in this list are forbidden to be chosen from. After
        randomly choose a permitted element, this will be added to the forbidden list and the first element of this
        list will be removed. This is a naive mechanism which guaranties the appropriate distance between two
        appearances of an element according for :obj:`periodicity`

        The ``forbidden_list`` can also be set manually. In this case if its length is larger than
        :obj:`periodicity`, :obj:`permutation_order_iterator` will remove so many elements from the beginning of this list until the
        right length is achieved.

        >>> r = MagicRandom(pool=[1, 3, 2, 4, 5, 6], periodicity=4, seed=20, forbidden_list=[2, 3, 1])
        >>> previous_forbidden_list = r.forbidden_list[:]
        >>> el1 = next(r)
        >>> el1 not in previous_forbidden_list
        True
        >>> r.forbidden_list == [2, 3, 1] + [el1]
        True
        >>> previous_forbidden_list = r.forbidden_list[:]
        >>> el2 = next(r)
        >>> el2 not in previous_forbidden_list
        True
        >>> r.forbidden_list == [3, 1] + [el1, el2]
        True
        >>> previous_forbidden_list = r.forbidden_list[:]
        >>> el3 = next(r)
        >>> el3 not in previous_forbidden_list
        True
        >>> r.forbidden_list == [1] + [el1, el2, el3]
        True
        """
        return self._forbidden_list

    @forbidden_list.setter
    def forbidden_list(self, values: Optional[list[Any]]) -> None:
        if not values:
            self._forbidden_list = []
        else:
            check_type(
                t=list,
                v=values,
                class_name=self.__class__.__name__,
                property_name="forbidden_list",
            )
            self._forbidden_list = values

    @property
    def periodicity(self) -> NonNegativeInteger:
        """
        Set and get ``periodicity`` property of types ``None`` or ``NonNegativInteger``. This property defines the
        minimum distance between two appearances of an element.
        If set to ``0`` immediate repetitions are permitted,
        if set to ``1`` at least one other element must be given out before this element can be chosen again an so on.
        If set to ``len(self.pool) - 1`` a random permutation of elements will be repeated.

        If set to ``None``, ``len(self.pool) - 2`` is returned. If len(self.pool) is ``1``, ``0`` is returned.
        If set to a value equal or greater than ``len(self.pool)``, ``len(self.pool) - 1`` is returned.

        >>> MagicRandom(pool=[1, 2, 3, 4]).periodicity
        2
        >>> MagicRandom(pool=[1]).periodicity
        0
        >>> MagicRandom(pool=[1, 2, 3, 4], periodicity=5).periodicity
        3
        """
        if self._periodicity is None:
            output = len(self.pool) - 2
            return max(output, 0)
        elif self._periodicity >= len(self.pool):
            return len(self.pool) - 1
        else:
            return self._periodicity

    @periodicity.setter
    def periodicity(self, value: Optional[NonNegativeInteger]) -> None:
        if value is None:
            self._periodicity = None
        else:
            check_type(
                t="NonNegativeInteger",
                v=value,
                property_name="periodicity",
                class_name=self.__class__.__name__,
            )
            self._periodicity = value

    @property
    def pool(self) -> list[Any]:
        """
        Set and get ``pool`` property. This property defines the list of possible elements to be randomly chosen from.
        Duplicates will be omitted without chaining the order of each element's first appearances.
        :return: ``None`` or ``list``

        >>> MagicRandom(pool=[1, 2, 3, 2, 1]).pool
        [1, 2, 3]
        """
        return self._pool

    @pool.setter
    def pool(self, values: list[Any]) -> None:
        check_type(
            v=values, t=list, property_name="pool", class_name=self.__class__.__name__
        )
        self._pool = list(dict.fromkeys(values))

    @property
    def seed(self) -> Optional[Union[int, str, bytes, bytearray]]:
        """
        Set and get ``seed.a`` value of python random function which is used to randomly choose an element.

        .. seealso:: https://docs.python.org/3/library/random.html#random.seed

        >>> MagicRandom(pool=[1, 2, 3], periodicity=0).seed

        >>> MagicRandom(pool=[1, 2, 3], periodicity=0, seed=10).seed
        10
        >>> MagicRandom(pool=[1, 2, 3], periodicity=0, seed='Can be a string too').seed
        'Can be a string too'
        """
        return self._seed

    @seed.setter
    def seed(self, value: Optional[Union[int, str, bytes, bytearray]]) -> None:
        self._seed = value
        self.current_random.seed(value)

    # methods
    def get_previous_elements(self) -> list[Any]:
        """
        :return: list of all randomly chosen values

        >>> r = MagicRandom(pool=[1, 3, 2, 4, 5], periodicity=2, seed=20)
        >>> [r.__next__() for _ in range(20)]
        [3, 2, 1, 5, 3, 1, 4, 3, 2, 4, 5, 3, 2, 4, 1, 5, 4, 1, 3, 5]
        >>> r.get_previous_elements()
        [3, 2, 1, 5, 3, 1, 4, 3, 2, 4, 5, 3, 2, 4, 1, 5, 4, 1, 3, 5]
        """
        return self._previous_elements

    def __iter__(self) -> Iterator[Any]:
        """
        The core methode of MagicRandom. This is a generator to generate random values. :obj:`__next__` method calls :obj:`__iter__().__next__()`.

        :return: a random value out of :obj:`pool` considering :obj:`seed`, :obj:`periodicity` and :obj:`forbidden_list`
        """
        while True:

            def check(element: Any) -> bool:
                def forbid_element(el: Any) -> None:
                    if len(self.forbidden_list) >= self.periodicity:
                        self.forbidden_list.pop(0)
                    self.forbidden_list.append(el)

                if self.periodicity != 0:
                    if element in self.forbidden_list:
                        return False
                    else:
                        forbid_element(element)
                        return True
                else:
                    return True

            if len(self.forbidden_list) > self.periodicity:
                self.forbidden_list = self.forbidden_list[(-1 * self.periodicity) :]

            random_element = self.pool[self.current_random.randrange(len(self.pool))]
            while check(random_element) is False:
                random_element = self.pool[
                    self.current_random.randrange(len(self.pool))
                ]
            self._counter += 1
            self._previous_elements.append(random_element)
            yield random_element

    def __next__(self) -> Any:
        """
        :return: self.__iter__().__next__()`
        """
        return self.__iter__().__next__()



from pathlib import Path

import matplotlib as mpl
import pytest

from musurgia.pdf.font import _make_afm_path_dictionary
from musurgia.tests.utils_for_tests import PdfTestCase

afm_path = Path(mpl.get_data_path(), "fonts", "afm", "ptmr8a.afm")


class Test(PdfTestCase):
    def test_afm_dict(self):
        actual = sorted(list(_make_afm_path_dictionary().keys()))
        expected = [
            ("Computer Modern", "medium", "italic"),
            ("Computer Modern", "medium", "regular"),
            ("Courier", "bold", "italic"),
            ("Courier", "bold", "regular"),
            ("Courier", "medium", "italic"),
            ("Courier", "medium", "regular"),
            ("Helvetica", "bold", "italic"),
            ("Helvetica", "bold", "regular"),
            ("Helvetica", "light", "italic"),
            ("Helvetica", "light", "regular"),
            ("Helvetica", "medium", "italic"),
            ("Helvetica", "medium", "regular"),
            ("ITC Avant Garde Gothic", "book", "italic"),
            ("ITC Avant Garde Gothic", "book", "regular"),
            ("ITC Avant Garde Gothic", "demi", "italic"),
            ("ITC Avant Garde Gothic", "demi", "regular"),
            ("ITC Bookman", "demi", "italic"),
            ("ITC Bookman", "demi", "regular"),
            ("ITC Bookman", "light", "italic"),
            ("ITC Bookman", "light", "regular"),
            ("ITC Zapf Chancery", "medium", "italic"),
            ("ITC Zapf Dingbats", "medium", "regular"),
            ("New Century Schoolbook", "bold", "italic"),
            ("New Century Schoolbook", "bold", "regular"),
            ("New Century Schoolbook", "medium", "italic"),
            ("New Century Schoolbook", "roman", "regular"),
            ("Palatino", "bold", "italic"),
            ("Palatino", "bold", "regular"),
            ("Palatino", "medium", "italic"),
            ("Palatino", "roman", "regular"),
            ("Symbol", "medium", "regular"),
            ("Times", "bold", "italic"),
            ("Times", "bold", "regular"),
            ("Times", "medium", "italic"),
            ("Times", "roman", "regular"),
            ("Utopia", "bold", "italic"),
            ("Utopia", "bold", "regular"),
            ("Utopia", "regular", "italic"),
            ("Utopia", "regular", "regular"),
        ]
        self.assertEqual(expected, actual)

    @pytest.mark.skip(reason="ci fails due to OS differences")
    def test_afm_dict_families(self):
        actual = list(
            dict.fromkeys([key[0] for key in _make_afm_path_dictionary().keys()])
        )
        expected = [
            "New Century Schoolbook",
            "Times",
            "ITC Bookman",
            "Helvetica",
            "ITC Avant Garde Gothic",
            "Palatino",
            "Computer Modern",
            "Symbol",
            "ITC Zapf Dingbats",
            "Utopia",
            "Courier",
            "ITC Zapf Chancery",
        ]
        self.assertEqual(expected, actual)

    def test_load_afm(self):
        afm = _make_afm_path_dictionary()["Times", "medium", "italic"]
        actual = afm.get_familyname()
        expected = "Times"
        self.assertEqual(expected, actual)

    @pytest.mark.skip(
        reason="ci fails due to OS differences: (7370.0, 741) != (6040.0, 741)"
    )
    def test_width_height(self):
        afm = _make_afm_path_dictionary()["Helvetica", "bold", "italic"]
        actual = afm.string_width_height("What the heck?")
        expected = (7370.0, 741)
        self.assertEqual(expected, actual)



from unittest import TestCase

from fractions import Fraction
import warnings

from musurgia.tests.utils_for_tests import DemoValuedTree, create_test_valued_tree
from musurgia.utils import flatten


class ValuedTreeChangeValueTestCase(TestCase):
    def setUp(self) -> None:
        self.vt = DemoValuedTree(value=10)

    def _add_layer(self, number_of_layers=1):
        for value in [5, Fraction(5, 3), Fraction(10, 3)]:
            self.vt.add_child(DemoValuedTree(value=value))
        if number_of_layers == 2:
            list_of_values = [
                [Fraction(5, 6), Fraction(5, 3), Fraction(5, 2)],
                [Fraction(5, 6), Fraction(5, 18), Fraction(5, 9)],
                [Fraction(10, 9), Fraction(5, 3), Fraction(5, 9)],
            ]
            for child, values in zip(self.vt.get_children(), list_of_values):
                for v in values:
                    child.add_child(DemoValuedTree(value=v))

    def test_change_root_value_without_children(self):
        self.vt.update_value(15)
        self.assertEqual(15, self.vt.get_value())
        self.vt.check_tree_values()

    def test_change_root_value_with_children(self):
        self._add_layer()
        self.vt.update_value(15)
        self.assertEqual(15, self.vt.get_value())
        self.assertEqual(
            15, sum([child.get_value() for child in self.vt.get_children()])
        )
        self.assertEqual(
            [Fraction(15, 2), Fraction(5, 2), Fraction(5, 1)],
            [child.get_value() for child in self.vt.get_children()],
        )
        self.vt.check_tree_values()

    def test_change_leaf_value(self):
        self._add_layer()
        self.vt.get_children()[0].update_value(10)
        self.assertEqual(15, self.vt.get_value())
        self.assertEqual(15, sum([child.get_value() for child in self.vt.get_layer(1)]))
        self.assertEqual(
            [Fraction(10, 1), Fraction(5, 3), Fraction(10, 3)],
            [child.get_value() for child in self.vt.get_children()],
        )
        self.vt.check_tree_values()

    def test_two_layers_change_child_value(self):
        self._add_layer(2)
        self.vt.get_children()[0].update_value(10)
        assert self.vt.get_value() == 15
        assert (
            sum(flatten(self.vt.get_layer(1, key=lambda node: node.get_value()))) == 15
        )
        assert (
            sum(flatten(self.vt.get_layer(2, key=lambda node: node.get_value()))) == 15
        )
        self.vt.check_tree_values()

    def test_with_remove(self):
        def node_info(node):
            return f"{node.get_position_in_tree()}: {round(float(node.get_value()), 2)}"

        self._add_layer()
        first_child = self.vt.get_children()[0]
        values = [Fraction(5, 6), Fraction(5, 3), Fraction(5, 2)]
        for v in values:
            first_child.add_child(DemoValuedTree(v))
        list_of_values = [
            [Fraction(5, 18), Fraction(5, 12), Fraction(5, 36)],
            [Fraction(5, 18), Fraction(5, 9), Fraction(5, 6)],
            [Fraction(5, 4), Fraction(5, 12), Fraction(5, 6)],
        ]
        for child, values in zip(first_child.get_children(), list_of_values):
            for v in values:
                child.add_child(DemoValuedTree(v))

        expected = """└── 0: 10.0
    ├── 1: 5.0
    │   ├── 1.1: 0.83
    │   │   ├── 1.1.1: 0.28
    │   │   ├── 1.1.2: 0.42
    │   │   └── 1.1.3: 0.14
    │   ├── 1.2: 1.67
    │   │   ├── 1.2.1: 0.28
    │   │   ├── 1.2.2: 0.56
    │   │   └── 1.2.3: 0.83
    │   └── 1.3: 2.5
    │       ├── 1.3.1: 1.25
    │       ├── 1.3.2: 0.42
    │       └── 1.3.3: 0.83
    ├── 2: 1.67
    └── 3: 3.33
"""
        assert self.vt.get_tree_representation(node_info) == expected
        first_child.remove(first_child._get_children()[1])
        expected = """└── 0: 10.0
    ├── 1: 5.0
    │   ├── 1.1: 0.83
    │   │   ├── 1.1.1: 0.28
    │   │   ├── 1.1.2: 0.42
    │   │   └── 1.1.3: 0.14
    │   └── 1.2: 2.5
    │       ├── 1.2.1: 1.25
    │       ├── 1.2.2: 0.42
    │       └── 1.2.3: 0.83
    ├── 2: 1.67
    └── 3: 3.33
"""
        with warnings.catch_warnings():
            assert self.vt.get_tree_representation(node_info) == expected
        first_child._get_children()[0].update_value(2.5)
        expected = """└── 0: 10.0
    ├── 1: 5.0
    │   ├── 1.1: 2.5
    │   │   ├── 1.1.1: 0.83
    │   │   ├── 1.1.2: 1.25
    │   │   └── 1.1.3: 0.42
    │   └── 1.2: 2.5
    │       ├── 1.2.1: 1.25
    │       ├── 1.2.2: 0.42
    │       └── 1.2.3: 0.83
    ├── 2: 1.67
    └── 3: 3.33
"""
        assert self.vt.get_tree_representation(node_info) == expected
        self.vt.check_tree_values()

    def test_update_child_check_root(self):
        vt = DemoValuedTree(60)
        for d in [20, 10, 30]:
            vt.add_child(DemoValuedTree(d))
        for list_of_d, child in zip(
            [[10, 2, 3, 5], [], [5, 20, 3, 2]], vt.get_children()
        ):
            for d in list_of_d:
                child.add_child(DemoValuedTree(d))
        expected = """└── 60.0
    ├── 20.0
    │   ├── 10.0
    │   ├── 2.0
    │   ├── 3.0
    │   └── 5.0
    ├── 10.0
    └── 30.0
        ├── 5.0
        ├── 20.0
        ├── 3.0
        └── 2.0
"""
        assert (
            vt.get_tree_representation(key=lambda node: float(node.get_value()))
            == expected
        )
        one_child = vt.get_children()[0].get_children()[0]
        one_child.update_value(15)
        vt.check_tree_values()
        expected = """└── 65.0
    ├── 25.0
    │   ├── 15.0
    │   ├── 2.0
    │   ├── 3.0
    │   └── 5.0
    ├── 10.0
    └── 30.0
        ├── 5.0
        ├── 20.0
        ├── 3.0
        └── 2.0
"""
        assert (
            vt.get_tree_representation(key=lambda node: float(node.get_value()))
            == expected
        )

    def test_update_child_of_util_valued_tree(self):
        vt = create_test_valued_tree()
        child = vt.get_children()[-1]
        assert child.get_value() == 2
        child.update_value(15)
        assert child.get_value() == 15
        assert vt.check_tree_values()


class WrongValueErrorTest(TestCase):
    def test_less_child_values(self):
        vt = DemoValuedTree(10)
        vt.add_child(DemoValuedTree(4))
        vt.add_child(DemoValuedTree(4))
        with warnings.catch_warnings(record=True) as w:
            vt.get_children()
            assert len(w) == 1

    def test_greate_child_values(self):
        vt = DemoValuedTree(10)
        vt.add_child(DemoValuedTree(4))
        vt.add_child(DemoValuedTree(4))
        vt.add_child(DemoValuedTree(4))
        with warnings.catch_warnings(record=True) as w:
            vt.get_children()
            assert len(w) == 1



from pathlib import Path
from typing import Any

from musurgia.musurgia_types import MarginType
from musurgia.pdf.labeled import Labeled, TextLabel
from musurgia.pdf.line import VerticalSegmentedLine
from musurgia.pdf.pdf import Pdf
from musurgia.pdf.pdf_tools import draw_ruler
from musurgia.pdf.drawobject import Master
from musurgia.tests.utils_for_tests import PdfTestCase

path = Path(__file__)


class LabeledVerticalLine(VerticalSegmentedLine, Labeled):
    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)

    def draw(self, pdf):
        self.draw_above_text_labels(pdf)
        self.draw_left_text_labels(pdf)
        super().draw(pdf)


class DummyPositionMaster(Master):
    def get_slave_position(self, slave, position):
        if position == "x":
            return 0
        elif position == "y":
            return 0

    def get_slave_margin(self, slave: Any, margin: MarginType) -> float:
        raise NotImplementedError  # pragma: no cover


class TestTextLabel(PdfTestCase):
    def setUp(self) -> None:
        self.pdf = Pdf()

    def test_get_text_labels(self):
        t = Labeled()
        assert t.get_text_labels() == []
        a = t.add_label(TextLabel("above", placement="above"))
        b = t.add_label(TextLabel("below", placement="below"))
        l = t.add_label(TextLabel("left", placement="left"))
        aa = t.add_label(TextLabel("above-above"))
        assert t.get_text_labels() == [l, a, aa, b]

    def test_draw(self):
        t = TextLabel(
            master=DummyPositionMaster(), name="t1", value="Fox is going to be dead."
        )

        with self.file_path(path, "draw", "pdf") as pdf_path:
            self.pdf.translate_page_margins()
            draw_ruler(self.pdf, "h")
            draw_ruler(self.pdf, "v")
            self.pdf.translate(10, 10)
            t.draw(self.pdf)
            self.pdf.write_to_path(pdf_path)

    def test_draw_multiple(self):
        t1 = TextLabel(
            master=DummyPositionMaster(), name="t1", value="Fox is going to be dead."
        )
        t1.top_margin = -5
        t2 = TextLabel(
            master=DummyPositionMaster(), name="t2", value="What should we do??"
        )
        t3 = TextLabel(
            master=DummyPositionMaster(), name="t3", value="What should we do??"
        )
        with self.file_path(path, "draw_multiple", "pdf") as pdf_path:
            self.pdf.translate_page_margins()
            draw_ruler(self.pdf, "h")
            draw_ruler(self.pdf, "v")
            self.pdf.translate(10, 10)
            t1.draw(self.pdf)
            self.pdf.translate(0, t1.get_height())
            t2.draw(self.pdf)
            self.pdf.translate(0, t2.get_height())
            t3.draw(self.pdf)
            self.pdf.translate(0, t3.get_height())
            self.pdf.write_to_path(pdf_path)

    def test_left_labels(self):
        unit = 1
        length = 40

        lvl = LabeledVerticalLine(lengths=int(length / unit) * [unit])
        for index, segment in enumerate(lvl.segments):
            if index % 10 == 0:
                segment.start_mark_line.length = 3
            else:
                segment.start_mark_line.length = 1.5

        lvl.add_label("first one one", placement="left", font_size=8)
        lvl.add_label("second two two", placement="left")
        lvl.add_label("third three three", placement="left", font_size=12)
        lvl.add_label("fourth four four", placement="left")
        lvl.add_label("fifth five five", placement="left")
        lvl.add_label("sixth six six", placement="left")
        lvl.add_label("seventh seven seven", placement="left")
        with self.file_path(path, "left_labels", "pdf") as pdf_path:
            self.pdf.translate_page_margins()
            draw_ruler(self.pdf, "h")
            draw_ruler(self.pdf, "v")
            self.pdf.translate(30, 10)
            lvl.draw(self.pdf)
            self.pdf.write_to_path(pdf_path)



from unittest import TestCase

from fractions import Fraction
import warnings

from musurgia.trees.fractaltimelinetree import FractalTimelineTree
from musurgia.tests.utils_for_tests import fractal_node_info
from musurgia.trees.timelinetree import TimelineDuration
from musurgia.utils import flatten


class Test(TestCase):
    def setUp(self) -> None:
        self.ft = FractalTimelineTree(
            duration=TimelineDuration(10),
            proportions=(1, 2, 3),
            main_permutation_order=(3, 1, 2),
            permutation_index=(1, 1),
        )

    def test_change_root_value_without_children(self):
        self.ft.update_value(15)
        self.assertEqual(15, self.ft.get_value())

    def test_change_root_value_with_children(self):
        self.ft.add_layer()
        self.ft.update_value(15)
        self.assertEqual(15, self.ft.get_value())
        self.assertEqual(
            15, sum([child.get_value() for child in self.ft.get_children()])
        )
        self.assertEqual(
            [Fraction(15, 2), Fraction(5, 2), Fraction(5, 1)],
            [child.get_value() for child in self.ft.get_children()],
        )

    def test_change_leaf_value(self):
        self.ft.add_layer()
        self.ft.get_children()[0].update_value(10)
        self.assertEqual(15, self.ft.get_value())
        self.assertEqual(15, sum([child.get_value() for child in self.ft.get_layer(1)]))
        self.assertEqual(
            [Fraction(10, 1), Fraction(5, 3), Fraction(10, 3)],
            [child.get_value() for child in self.ft.get_children()],
        )

    def test_two_layers_change_child_value(self):
        self.ft.add_layer()
        self.ft.add_layer()
        self.ft.get_children()[0].update_value(10)
        assert self.ft.get_value() == 15
        assert (
            sum(flatten(self.ft.get_layer(1, key=lambda node: node.get_value()))) == 15
        )
        assert (
            sum(flatten(self.ft.get_layer(2, key=lambda node: node.get_value()))) == 15
        )

    def test_with_remove(self):
        self.ft.add_layer()
        first_child = self.ft.get_children()[0]
        first_child.add_layer()
        first_child.add_layer()
        # print(self.ft.get_tree_representation(node_info))
        assert (
            self.ft.get_tree_representation(fractal_node_info)
            == """└── 0: (1, 1): 10.0
    ├── 3: (2, 1): 5.0
    │   ├── 1: (3, 1): 0.83
    │   │   ├── 2: (1, 1): 0.28
    │   │   ├── 3: (1, 2): 0.42
    │   │   └── 1: (1, 3): 0.14
    │   ├── 2: (3, 2): 1.67
    │   │   ├── 1: (2, 1): 0.28
    │   │   ├── 2: (2, 2): 0.56
    │   │   └── 3: (2, 3): 0.83
    │   └── 3: (3, 3): 2.5
    │       ├── 3: (3, 1): 1.25
    │       ├── 1: (3, 2): 0.42
    │       └── 2: (3, 3): 0.83
    ├── 1: (2, 2): 1.67
    └── 2: (2, 3): 3.33
"""
        )
        first_child.remove(first_child.get_children()[1])
        # print(self.ft.get_tree_representation(node_info))
        with warnings.catch_warnings():
            assert (
                self.ft.get_tree_representation(fractal_node_info)
                == """└── 0: (1, 1): 10.0
    ├── 3: (2, 1): 5.0
    │   ├── 1: (3, 1): 0.83
    │   │   ├── 2: (1, 1): 0.28
    │   │   ├── 3: (1, 2): 0.42
    │   │   └── 1: (1, 3): 0.14
    │   └── 3: (3, 3): 2.5
    │       ├── 3: (3, 1): 1.25
    │       ├── 1: (3, 2): 0.42
    │       └── 2: (3, 3): 0.83
    ├── 1: (2, 2): 1.67
    └── 2: (2, 3): 3.33
"""
            )
        first_child._get_children()[0].update_value(2.5)
        # print(self.ft.get_tree_representation(node_info))
        assert (
            self.ft.get_tree_representation(fractal_node_info)
            == """└── 0: (1, 1): 10.0
    ├── 3: (2, 1): 5.0
    │   ├── 1: (3, 1): 2.5
    │   │   ├── 2: (1, 1): 0.83
    │   │   ├── 3: (1, 2): 1.25
    │   │   └── 1: (1, 3): 0.42
    │   └── 3: (3, 3): 2.5
    │       ├── 3: (3, 1): 1.25
    │       ├── 1: (3, 2): 0.42
    │       └── 2: (3, 3): 0.83
    ├── 1: (2, 2): 1.67
    └── 2: (2, 3): 3.33
"""
        )



from musurgia.pdf.font import Font
from musurgia.tests.utils_for_tests import PdfTestCase


class TestFont(PdfTestCase):
    def setUp(self) -> None:
        self.font = Font()

    def test_load_afm(self):
        afm = self.font._afm
        actual = afm.get_familyname()
        expected = self.font.family
        self.assertEqual(expected, actual)

    def test_text_pixel_height(self):
        self.font.size = 12
        actual = round(self.font.get_text_pixel_height("This One"), 3)
        expected = 8.1
        self.assertEqual(expected, actual)

    def test_text_pixel_width(self):
        self.font.size = 11
        actual = round(self.font.get_text_pixel_width("This One"), 3)
        expected = 52.8
        self.assertEqual(expected, actual)

    def test_errors(self):
        with self.assertRaises(TypeError):
            self.font.family = "Bla"

        with self.assertRaises(TypeError):
            self.font.style = "Bla"

        with self.assertRaises(TypeError):
            self.font.weight = "Bla"



from pathlib import Path

from musurgia.pdf.line import VerticalSegmentedLine
from musurgia.pdf.pdf import Pdf
from musurgia.pdf.pdf_tools import draw_ruler, draw_page_numbers
from musurgia.pdf.text import Text
from musurgia.tests.utils_for_tests import PdfTestCase

path = Path(__file__)


class TestAddPage(PdfTestCase):
    def setUp(self) -> None:
        self.pdf = Pdf()

    def test_absolute_y(self):
        self.pdf.translate(10, 10)
        self.pdf.add_page()
        actual = self.pdf.absolute_positions
        expected = {1: [10, 10], 2: [0, 0]}
        self.assertEqual(expected, actual)

    def test_draw(self):
        self.pdf.translate_page_margins()
        draw_ruler(self.pdf, mode="h")
        draw_ruler(self.pdf, mode="v")
        self.pdf.translate(10, 10)
        vsl = VerticalSegmentedLine(lengths=[10, 20, 30, 40])
        vsl.draw(self.pdf)
        self.pdf.add_page()
        self.pdf.translate_page_margins()
        draw_ruler(self.pdf, mode="h")
        draw_ruler(self.pdf, mode="v")
        self.pdf.translate(10, 10)
        # print(self.pdf.absolute_positions)
        vsl.draw(self.pdf)
        with self.file_path(path, "draw", "pdf") as pdf_path:
            self.pdf.write_to_path(pdf_path)

    def test_page_number(self):
        for i in range(3):
            self.pdf.add_page()
        draw_page_numbers(self.pdf, v_position="bottom", h_position="center")

        with self.file_path(path, "page_number", "pdf") as pdf_path:
            self.pdf.write_to_path(pdf_path)

    def test_draw_in_reversed_order(self):
        self.pdf.translate(30, 30)
        self.pdf.add_page()
        self.pdf.translate(80, 80)
        self.pdf.add_page()
        self.pdf.page = 2
        text = Text("second page")
        text.draw(self.pdf)
        self.pdf.page = 1
        text = Text("first page")
        text.draw(self.pdf)

        for page in self.pdf.pages:
            self.pdf.page = page
            self.pdf.reset_position()
            self.pdf.translate_page_margins()
            draw_ruler(self.pdf, "h")
            draw_ruler(self.pdf, "v")

        draw_page_numbers(self.pdf, v_position="bottom", h_position="center")

        with self.file_path(path, "draw_in_reversed_order", "pdf") as pdf_path:
            self.pdf.write_to_path(pdf_path)



from musurgia.magicrandom import MagicRandom
from musurgia.tests.utils_for_tests import PdfTestCase


class TestRandom(PdfTestCase):
    def test_counter(self):
        r = MagicRandom(pool=["a", "b", "c"], periodicity=1, seed=20)
        assert r.counter == 0
        assert next(r) == "c"

        assert r.counter == 1
        assert next(r) == "a"
        assert r.counter == 2

    def test_forbidden_list(self):
        with self.assertRaises(TypeError):
            MagicRandom(pool=[1, 2, 3], forbidden_list="[2]")

        r = MagicRandom(
            pool=[1, 3, 2, 4, 5, 6], periodicity=4, seed=20, forbidden_list=[2, 3, 1]
        )
        previous_forbidden_list = r.forbidden_list[:]
        el1 = next(r)
        assert el1 not in previous_forbidden_list
        assert r.forbidden_list == [2, 3, 1] + [el1]
        previous_forbidden_list = r.forbidden_list[:]
        el2 = next(r)
        assert r not in previous_forbidden_list
        assert r.forbidden_list == [3, 1] + [el1, el2]
        previous_forbidden_list = r.forbidden_list[:]
        el3 = next(r)
        assert el3 not in previous_forbidden_list
        assert r.forbidden_list == [1] + [el1, el2, el3]

    def test_periodicity(self):
        with self.assertRaises(TypeError):
            MagicRandom(pool=[1, 2, 3], periodicity=-1)
        with self.assertRaises(TypeError):
            MagicRandom(pool=[1, 2, 3], periodicity=1.6)

        assert MagicRandom(pool=[1, 2, 3, 4]).periodicity == 2
        assert MagicRandom(pool=[1]).periodicity == 0
        assert MagicRandom(pool=[1, 2, 3, 4], periodicity=5).periodicity == 3

    def test_pool(self):
        with self.assertRaises(TypeError):
            MagicRandom(pool=3)
        with self.assertRaises(TypeError):
            MagicRandom(pool=None)
        assert MagicRandom(pool=[1, 2, 3, 2, 1]).pool == [1, 2, 3]

    def test_seed(self):
        assert MagicRandom(pool=[1, 2, 3], periodicity=0).seed is None
        assert MagicRandom(pool=[1, 2, 3], periodicity=0, seed=10).seed == 10
        assert (
            MagicRandom(pool=[1, 2, 3], periodicity=0, seed="Can be a string too").seed
            == "Can be a string too"
        )

    def test_get_previous_elements(self):
        r = MagicRandom(pool=[1, 3, 2, 4, 5], periodicity=2, seed=20)
        assert [r.__next__() for _ in range(20)] == [
            3,
            2,
            1,
            5,
            3,
            1,
            4,
            3,
            2,
            4,
            5,
            3,
            2,
            4,
            1,
            5,
            4,
            1,
            3,
            5,
        ]

        assert r.get_previous_elements() == [
            3,
            2,
            1,
            5,
            3,
            1,
            4,
            3,
            2,
            4,
            5,
            3,
            2,
            4,
            1,
            5,
            4,
            1,
            3,
            5,
        ]



from typing import Union, Any

from musurgia.musurgia_types import check_type, HorizontalVertical
from musurgia.pdf.ruler import HorizontalRuler, VerticalRuler
from musurgia.pdf.pdf import Pdf
from musurgia.pdf.text import PageText

__all__ = ["draw_page_numbers", "draw_ruler"]


def draw_page_numbers(pdf: Pdf, **kwargs: Any) -> None:
    for page in pdf.pages:
        pdf.page = page
        pdf.reset_position()
        page_number = PageText(page, **kwargs)
        page_number.draw(pdf)


def draw_ruler(
    pdf: Pdf,
    mode: HorizontalVertical = "h",
    unit: int = 10,
    first_label: int = 0,
    show_first_label: bool = False,
    label_show_interval: int = 1,
    **kwargs: Any,
) -> None:
    check_type(
        mode, "HorizontalVertical", function_name="draw_ruler", argument_name="mode"
    )
    ruler: Union[HorizontalRuler, VerticalRuler]
    if mode in ["h", "horizontal"]:
        length = pdf.w - pdf.l_margin - pdf.r_margin
        ruler = HorizontalRuler(
            length=length,
            unit=unit,
            first_label=first_label,
            show_first_label=show_first_label,
            label_show_interval=label_show_interval,
            **kwargs,
        )
        ruler.set_straight_line_relative_y(0)
    elif mode in ["v", "vertical"]:
        length = pdf.h - pdf.t_margin - pdf.b_margin
        ruler = VerticalRuler(
            length=length,
            unit=unit,
            first_label=first_label,
            show_first_label=show_first_label,
            label_show_interval=label_show_interval,
            **kwargs,
        )
        ruler.set_straight_line_relative_x(0)
    else:
        raise NotImplementedError  # pragma: no cover

    ruler.draw(pdf)
